import numpy as np

from .analysis_synthesis import AnalysisSynthesisTSM
from .utils.windows import hanning


def find_peaks(amplitude):
    """
    A value is considered to be a peak if it is higher than its four closest
    neighbours.
    """

    # Pad the array with -1 at the beginning and the end to avoid overflows.
    padded = np.concatenate((-np.ones(2), amplitude, -np.ones(2)))

    # Shift the array by one/two values to the left/right
    shifted_l2 = padded[:-4]
    shifted_l1 = padded[1:-3]
    shifted_r1 = padded[3:-1]
    shifted_r2 = padded[4:]

    # Compare the original array with the shifted versions.
    peaks = (
        (amplitude >= shifted_l2)
        & (amplitude >= shifted_l1)
        & (amplitude >= shifted_r1)
        & (amplitude >= shifted_r2)
    )

    return peaks


def get_closest_peaks(peaks):
    """
    Returns an array containing the index of the closest peak of each index.
    """
    closest_peak = np.empty_like(peaks, dtype=int)
    previous = -1
    for i, is_peak in enumerate(peaks):
        if is_peak:
            if previous >= 0:
                closest_peak[previous : (previous + i) // 2 + 1] = previous
                closest_peak[(previous + i) // 2 + 1 : i] = i
            else:
                closest_peak[:i] = i
            previous = i
    closest_peak[previous:] = previous

    return closest_peak


class PhaseVocoderConverter:
    """
    A Converter implementing the phase vocoder time-scale modification procedure.
    """

    def __init__(
        self, channels, frame_length, analysis_hop, synthesis_hop, peak_finder
    ):
        self._channels = channels
        self._frame_length = frame_length
        self._synthesis_hop = synthesis_hop
        self._analysis_hop = analysis_hop
        self._find_peaks = peak_finder

        # Centers of the FFT frequency bins
        self._center_frequency = np.fft.rfftfreq(frame_length) * 2 * np.pi
        fft_length = len(self._center_frequency)

        self._first = True

        self._previous_phase = np.empty((channels, fft_length))
        self._output_phase = np.empty((channels, fft_length))

        # Buffer used to compute the phase increment and the instantaneous frequency
        self._buffer = np.empty(fft_length)

    def clear(self):
        self._first = True

    def convert_frame(self, frame):
        for k in range(0, self._channels):
            # Compute the FFT of the analysis frame
            stft = np.fft.rfft(frame[k])
            amplitude = np.abs(stft)
            phase = np.angle(stft)
            del stft

            peaks = self._find_peaks(amplitude)
            closest_peak = get_closest_peaks(peaks)

            if self._first:
                # Leave the first frame unchanged
                self._output_phase[k, :] = phase
            else:
                # Compute the phase increment
                self._buffer[peaks] = (
                    phase[peaks]
                    - self._previous_phase[k, peaks]
                    - self._analysis_hop * self._center_frequency[peaks]
                )

                # Unwrap the phase increment
                self._buffer[peaks] += np.pi
                self._buffer[peaks] %= 2 * np.pi
                self._buffer[peaks] -= np.pi

                # Compute the instantaneous frequency (in the same buffer,
                # since the phase increment wont be required after that)
                self._buffer[peaks] /= self._analysis_hop
                self._buffer[peaks] += self._center_frequency[peaks]

                self._buffer[peaks] *= self._synthesis_hop
                self._output_phase[k][peaks] += self._buffer[peaks]

                # Phase locking
                self._output_phase[k] = (
                    self._output_phase[k][closest_peak] + phase - phase[closest_peak]
                )

                # Compute the new stft
                output_stft = amplitude * np.exp(1j * self._output_phase[k])

                frame[k, :] = np.fft.irfft(output_stft).real

            # Save the phase for the next analysis frame
            self._previous_phase[k, :] = phase
            del phase
            del amplitude

        self._first = False

        return frame

    def set_analysis_hop(self, analysis_hop):
        self._analysis_hop = analysis_hop


def phasevocoder(
    channels, speed=1.0, frame_length=2048, analysis_hop=None, synthesis_hop=None
):

    if synthesis_hop is None:
        synthesis_hop = frame_length // 4

    if analysis_hop is None:
        analysis_hop = int(synthesis_hop * speed)

    analysis_window = hanning(frame_length)
    synthesis_window = hanning(frame_length)

    peak_finder = find_peaks

    converter = PhaseVocoderConverter(
        channels, frame_length, analysis_hop, synthesis_hop, peak_finder
    )

    return AnalysisSynthesisTSM(
        converter,
        channels,
        frame_length,
        analysis_hop,
        synthesis_hop,
        analysis_window,
        synthesis_window,
    )



# type: ignore

from dataclasses import dataclass

from auto_editor.utils.types import (
    float_type,
    anchor_type,
    color_type,
    text_content,
    align_type,
)


@dataclass
class TimelineObject:
    start: int
    dur: int


@dataclass
class RectangleObject(TimelineObject):
    x: int
    y: int
    width: int
    height: int
    anchor: anchor_type = "ce"
    fill: color_type = "#c4c4c4"
    stroke: int = 0
    strokecolor: color_type = "#000"
    _type: str = "rectangle"


@dataclass
class EllipseObject(TimelineObject):
    x: int
    y: int
    width: int
    height: int
    anchor: anchor_type = "ce"
    fill: color_type = "#c4c4c4"
    stroke: int = 0
    strokecolor: color_type = "#000"
    _type: str = "ellipse"


@dataclass
class TextObject(TimelineObject):
    content: text_content
    x: int = "centerX"
    y: int = "centerY"
    size: int = 30
    font: str = "default"
    align: align_type = "left"
    fill: color_type = "#000"
    stroke: int = 0
    strokecolor: color_type = "#000"
    _type: str = "text"


@dataclass
class ImageObject(TimelineObject):
    src: str
    x: int = "centerX"
    y: int = "centerY"
    opacity: float_type = 1
    anchor: anchor_type = "ce"
    rotate: float = 0  # in degrees
    _type: str = "image"



import numpy as np


def apply(buffer, window):
    """
    Applies a window to a buffer.
    """
    if window is None:
        return

    for channel in buffer:
        channel *= window


def hanning(length):
    """
    Returns a periodic Hanning window.
    """
    if length <= 0:
        return np.zeros(0)

    time = np.arange(length)
    return 0.5 * (1 - np.cos(2 * np.pi * time / length))


def product(window1, window2):
    if window1 is None:
        return window2

    if window2 is None:
        return window1

    return window1 * window2



"""
Make a pre-edited file reference that can be inputted back into auto-editor.
"""

import os
import sys
import json

from typing import List, Tuple, Union

from auto_editor.utils.log import Log


def check_attrs(data, log: Log, *attrs: str):
    for attr in attrs:
        if attr not in data:
            log.error(f"'{attr}' attribute not found!")


def check_file(path: str, log: Log):
    if not os.path.isfile(path):
        log.error(f"Could not locate media file: '{path}'")


def validate_chunks(chunks, log: Log) -> List[Tuple[int, int, float]]:

    if len(chunks) == 0:
        log.error("Chunks are empty!")

    new_chunks = []

    prev_end = None

    for i, chunk in enumerate(chunks):

        if len(chunk) != 3:
            log.error("Chunk has incorrect length.")

        if i == 0 and chunk[0] != 0:
            log.error("First chunk must start with 0")

        if chunk[1] - chunk[0] < 1:
            log.error("Chunk duration must be at least 1")

        if chunk[2] <= 0 or chunk[2] > 99999:
            log.error("Chunk speed range must be >0 and <=99999")

        if prev_end is not None and chunk[0] != prev_end:
            log.error(f"Chunk disjointed at {chunk}")

        prev_end = chunk[1]

        new_chunks.append((chunk[0], chunk[1], float(chunk[2])))

    return new_chunks


def read_json_timeline(json_file: str, log: Log):
    with open(json_file, "r") as f:
        data = json.load(f)

    check_attrs(data, log, "version")

    if data["version"] not in ("0.1.0", "0.2.0"):
        log.error(f"Unsupported version: {data['version']}")

    if data["version"] == "0.1.0":
        check_attrs(data, log, "source", "chunks")
        check_file(data["source"], log)

        chunks = validate_chunks(data["chunks"], log)

        return "#000", data["source"], chunks

    # version 0.2.0
    check_attrs(data, log, "source", "background", "chunks", "timeline")
    check_file(data["source"], log)

    chunks = validate_chunks(data["chunks"], log)

    return data["background"], data["source"], chunks


def make_json_timeline(
    version: str,
    media_file: str,
    out: Union[str, int],
    obj_sheet,
    chunks: List[Tuple[int, int, float]],
    background,
    log: Log,
) -> None:

    if version not in ("0.1.0", "0.2.0"):
        log.error(f"Version {version} is not supported!")

    if version == "0.1.0":
        data = {
            "version": "0.1.0",
            "source": os.path.abspath(media_file),
            "chunks": chunks,
        }
    if version == "0.2.0":
        data = {
            "version": "0.2.0",
            "source": os.path.abspath(media_file),
            "background": background,
            "chunks": chunks,
            "timeline": obj_sheet.all,
        }

    if isinstance(out, str):
        if not out.endswith(".json"):
            log.error("Output extension must be .json")

        with open(out, "w") as outfile:
            json.dump(data, outfile, indent=2, default=lambda o: o.__dict__)
    else:
        json.dump(data, sys.stdout, indent=2, default=lambda o: o.__dict__)
        print("")  # Flush stdout



# Internal Libraries
import re
import sys
import os.path
import subprocess
from platform import system
from dataclasses import dataclass
from subprocess import Popen, PIPE

# Typing
from typing import List, Optional

# Included Libraries
from auto_editor.utils.func import get_stdout
from auto_editor.utils.log import Log


def regex_match(regex: str, text: str) -> Optional[str]:
    match = re.search(regex, text)
    if match:
        return match.groupdict()["match"]
    return None


class FFmpeg:
    __slots__ = ("debug", "path", "version")

    @staticmethod
    def _set_ff_path(ff_location: Optional[str], my_ffmpeg: bool) -> str:
        if ff_location is not None:
            return ff_location
        if my_ffmpeg or system() not in ("Windows", "Darwin"):
            return "ffmpeg"
        program = "ffmpeg" if system() == "Darwin" else "ffmpeg.exe"
        dirpath = os.path.dirname(os.path.realpath(__file__))
        return os.path.join(dirpath, "ffmpeg", system(), program)

    def __init__(
        self,
        ff_location: Optional[str] = None,
        my_ffmpeg: bool = False,
        debug: bool = False,
    ) -> None:

        self.debug = debug
        self.path = self._set_ff_path(ff_location, my_ffmpeg)
        try:
            _version = get_stdout([self.path, "-version"]).split("\n")[0]
            _version = _version.replace("ffmpeg version", "").strip()
            self.version = _version.split(" ")[0]
        except FileNotFoundError:
            if system() == "Darwin":
                Log().error(
                    "No ffmpeg found, download via homebrew or restore the "
                    "included binary."
                )
            if system() == "Windows":
                Log().error(
                    "No ffmpeg found, download ffmpeg with your favorite package "
                    "manager (ex chocolatey), or restore the included binary."
                )

            Log().error("ffmpeg must be installed and on PATH.")

    def print(self, message: str) -> None:
        if self.debug:
            print(f"FFmpeg: {message}", file=sys.stderr)

    def print_cmd(self, cmd: List[str]) -> None:
        if self.debug:
            print(f"FFmpeg run: {' '.join(cmd)}\n", file=sys.stderr)

    def run(self, cmd: List[str]) -> None:
        cmd = [self.path, "-y", "-hide_banner"] + cmd
        if not self.debug:
            cmd.extend(["-nostats", "-loglevel", "error"])
        self.print_cmd(cmd)
        subprocess.run(cmd)

    def run_check_errors(
        self,
        cmd: List[str],
        log: Log,
        show_out: bool = False,
        path: Optional[str] = None,
    ) -> None:
        def _run(cmd: List[str]) -> str:
            process = self.Popen(cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE)

            _, stderr = process.communicate()

            if process.stdin is not None:
                process.stdin.close()
            return stderr.decode("utf-8", "replace")

        output = _run(cmd)

        if "Try -allow_sw 1" in output:
            cmd.insert(-1, "-allow_sw")
            cmd.insert(-1, "1")
            output = _run(cmd)

        error_list = [
            r"Unknown encoder '.*'",
            r"-q:v qscale not available for encoder\. Use -b:v bitrate instead\.",
            r"Specified sample rate .* is not supported",
            r'Unable to parse option value ".*"',
            r"Error setting option .* to value .*\.",
            r"Undefined constant or missing '.*' in '.*'",
            r"DLL .* failed to open",
            r"Incompatible pixel format '.*' for codec '[A-Za-z0-9_]*'",
            r"Unrecognized option '.*'",
            r"Permission denied",
        ]

        if self.debug:
            print(f"stderr: {output}")

        for item in error_list:
            check = re.search(item, output)
            if check:
                log.error(check.group())

        if path is not None and not os.path.isfile(path):
            print(output)
            log.error(f"The file {path} was not created.")
        elif show_out and not self.debug:
            print(f"stderr: {output}")

    def Popen(self, cmd: List[str], stdin=None, stdout=PIPE, stderr=None) -> Popen:
        cmd = [self.path] + cmd
        self.print_cmd(cmd)
        return Popen(cmd, stdin=stdin, stdout=stdout, stderr=stderr)

    def pipe(self, cmd: List[str]) -> str:
        cmd = [self.path, "-y"] + cmd

        self.print_cmd(cmd)
        output = get_stdout(cmd)
        self.print(output)
        return output


@dataclass
class VideoStream:
    width: int
    height: int
    codec: str
    fps: float
    bitrate: Optional[str]
    lang: Optional[str]


@dataclass
class AudioStream:
    codec: str
    samplerate: int
    bitrate: Optional[str]
    lang: Optional[str]


@dataclass
class SubtitleStream:
    codec: str
    ext: str
    lang: Optional[str]


class FileInfo:
    __slots__ = (
        "path",
        "abspath",
        "basename",
        "dirname",
        "name",
        "ext",
        "duration",
        "bitrate",
        "metadata",
        "videos",
        "audios",
        "subtitles",
        "gfps",
        "gwidth",
        "gheight",
        "gsamplerate",
    )

    def __init__(self, path: str, ffmpeg: FFmpeg, log: Log):
        self.path = path
        self.abspath = os.path.abspath(path)
        self.basename = os.path.basename(path)
        self.dirname = os.path.dirname(os.path.abspath(path))
        self.name, self.ext = os.path.splitext(path)

        info = get_stdout([ffmpeg.path, "-hide_banner", "-i", path])

        self.duration = regex_match(r"Duration:\s(?P<match>[0-9:.]+),", info)
        self.bitrate = regex_match(r"bitrate:\s(?P<match>[0-9]+\skb\/s)", info)

        self.metadata = {}
        active = False
        active_key = None

        for line in info.split("\n"):
            if active:
                if re.search(r"^\s*[A-Z][a-z_]*", line):
                    break

                key = regex_match(r"^\s*(?P<match>[a-z_]+)", line)
                body = regex_match(r"^\s*[a-z_]*\s*:\s(?P<match>[\w\W]*)", line)

                if body is not None:
                    if key is None:
                        if active_key is not None:
                            self.metadata[active_key] += "\n" + body
                    else:
                        self.metadata[key] = body
                        active_key = key

            if re.search(r"^\s\sMetadata:", line):
                active = True

        fps = None
        lang_regex = r"Stream #[0-9]+:[0-9](?:[\[\]a-z0-9]|)+\((?P<match>\w+)\)"
        sub_exts = {"mov_text": "srt", "ass": "ass", "webvtt": "vtt"}

        self.videos = []
        self.audios = []
        self.subtitles = []

        for line in info.split("\n"):
            if re.search(r"Stream #", line):
                if re.search(r"Video:", line):
                    codec = regex_match(r"Video:\s(?P<match>\w+)", line)
                    if codec is None:
                        log.error("Auto-Editor got 'None' when probing video codec")

                    w = regex_match(r"(?P<match>[0-9]+)x[0-9]+[\s,]", line)
                    h = regex_match(r"[0-9]+x(?P<match>[0-9]+)[\s,]", line)
                    _fps = regex_match(r"\s(?P<match>[0-9\.]+)\stbr", line)

                    if w is None or h is None:
                        log.error("Auto-Editor got 'None' when probing resolution")
                    if _fps is None:
                        if codec not in ("png", "mjpeg", "webp"):
                            log.error("Auto-Editor got 'None' when probing fps")
                        _fps = "30"

                    try:
                        fps = float(_fps)
                    except ValueError:
                        log.error(f"Couldn't convert '{fps}' to float")

                    if fps < 1:
                        log.error(
                            f"{self.basename}: Frame rate cannot be below 1. fps: {fps}"
                        )

                    self.videos.append(
                        VideoStream(
                            int(w),
                            int(h),
                            codec,
                            fps,
                            bitrate=regex_match(r"\s(?P<match>[0-9]+\skb\/s)", line),
                            lang=regex_match(lang_regex, line),
                        )
                    )
                elif re.search(r"Audio:", line):
                    _sr = regex_match(r"(?P<match>[0-9]+)\sHz", line)
                    if _sr is None:
                        log.error("Auto-Editor got 'None' when probing samplerate")

                    codec = regex_match(r"Audio:\s(?P<match>\w+)", line)
                    if codec is None:
                        log.error("Auto-Editor got 'None' when probing audio codec")

                    self.audios.append(
                        AudioStream(
                            codec,
                            int(_sr),
                            bitrate=regex_match(r"\s(?P<match>[0-9]+\skb\/s)", line),
                            lang=regex_match(lang_regex, line),
                        )
                    )
                elif re.search(r"Subtitle:", line):
                    codec = regex_match(r"Subtitle:\s(?P<match>\w+)", line)
                    if codec is None:
                        log.error("Auto-Editor got 'None' when probing subtitle codec")

                    ext = sub_exts.get(codec, "vtt")

                    self.subtitles.append(
                        SubtitleStream(codec, ext, regex_match(lang_regex, line))
                    )

        self.gfps: float = 30.0 if fps is None else fps

        if len(self.videos) > 0:
            self.gwidth, self.gheight = self.videos[0].width, self.videos[0].height
        else:
            self.gwidth, self.gheight = 1280, 720

        if len(self.audios) > 0:
            self.gsamplerate = self.audios[0].samplerate
        else:
            self.gsamplerate = 48000



from typing import Optional, List, Dict, Any

from dataclasses import dataclass, field

pcm_formats = [
    "pcm_s16le",  # default format
    "pcm_alaw",
    "pcm_f32be",
    "pcm_f32le",
    "pcm_f64be",
    "pcm_f64le",
    "pcm_mulaw",
    "pcm_s16be",
    "pcm_s24be",
    "pcm_s24le",
    "pcm_s32be",
    "pcm_s32le",
    "pcm_s8",
    "pcm_u16be",
    "pcm_u16le",
    "pcm_u24be",
    "pcm_u24le",
    "pcm_u32be",
    "pcm_u32le",
    "pcm_u8",
    "pcm_vidc",
]

# Define aliases
h265 = {
    "name": "H.265 / High Efficiency Video Coding (HEVC) / MPEG-H Part 2",
    "allow_video": True,
    "vcodecs": ["hevc", "mpeg4", "h264"],
}
h264 = {
    "name": "H.264 / Advanced Video Coding (AVC) / MPEG-4 Part 10",
    "allow_video": True,
    "vcodecs": ["h264", "mpeg4", "hevc"],
}
aac = {
    "name": "Advanced Audio Coding",
    "allow_audio": True,
    "max_audios": 1,
    "acodecs": ["aac"],
    "astrict": True,
}
ass = {
    "name": "SubStation Alpha",
    "allow_subtitle": True,
    "scodecs": ["ass", "ssa"],
    "max_subtitles": 1,
    "sstrict": True,
}
mp4 = {
    "name": "MP4 / MPEG-4 Part 14",
    "allow_video": True,
    "allow_audio": True,
    "allow_subtitle": True,
    "allow_image": True,
    "vcodecs": ["h264", "hevc", "vp9", "av1", "mpeg4", "mpeg2video", "mjpeg"],
    "acodecs": ["aac", "mp3", "opus", "flac", "vorbis", "mp2"],
    "disallow_v": ["prores", "apng", "gif", "msmpeg4v3", "flv1", "vp8", "rawvideo"],
    "disallow_a": pcm_formats,
}
ogg = {
    "allow_video": True,
    "allow_audio": True,
    "allow_subtitle": True,
    "vcodecs": ["libtheora", "theora"],
    "acodecs": ["libvorbis", "vorbis", "flac", "opus", "speex"],
    "vstrict": True,
    "astrict": True,
}

containers: Dict[str, Dict[str, Any]] = {
    # Aliases section
    "aac": aac,
    "adts": aac,
    "ass": ass,
    "ssa": ass,
    "264": h264,
    "h264": h264,
    "265": h265,
    "h265": h265,
    "hevc": h265,
    "mp4": mp4,
    "m4a": mp4,
    "ogg": ogg,
    "ogv": ogg,
    "apng": {
        "name": "Animated Portable Network Graphics",
        "allow_video": True,
        "max_videos": 1,
        "vcodecs": ["apng"],
        "vstrict": True,
    },
    "gif": {
        "name": "Graphics Interchange Format",
        "allow_video": True,
        "max_videos": 1,
        "vcodecs": ["gif"],
        "vstrict": True,
    },
    "wav": {
        "name": "Waveform Audio File Format",
        "allow_audio": True,
        "max_audios": 1,
        "acodecs": pcm_formats + ["mp3"],
        "astrict": True,
    },
    "ast": {
        "name": "AST / Audio Stream",
        "allow_audio": True,
        "acodecs": ["pcm_s16be_planar"],
    },
    "mp3": {
        "name": "MP3 / MPEG-2 Audio Layer 3",
        "allow_audio": True,
        "max_audios": 1,
        "acodecs": ["mp3"],
        "astrict": True,
    },
    "opus": {
        "name": "Opus",
        "allow_audio": True,
        "acodecs": ["opus", "flac", "libvorbis", "vorbis", "speex"],
        "astrict": True,
    },
    "oga": {
        "allow_audio": True,
        "acodecs": ["flac", "libvorbis", "vorbis", "opus", "speex"],
        "astrict": True,
    },
    "flac": {
        "name": "Free Lossless Audio Codec",
        "allow_audio": True,
        "max_audios": 1,
        "acodecs": ["flac"],
    },
    "webm": {
        "name": "WebM",
        "allow_video": True,
        "allow_audio": True,
        "allow_subtitle": True,
        "vcodecs": ["vp9", "vp8", "av1", "libaom-av1"],
        "acodecs": ["opus", "vorbis"],
        "scodecs": ["webvtt"],
        "vstrict": True,
        "astrict": True,
        "sstrict": True,
    },
    "srt": {
        "name": "SubRip Text / Subtitle Resource Tracks",
        "allow_subtitle": True,
        "scodecs": ["srt"],
        "max_subtitles": 1,
        "sstrict": True,
    },
    "vtt": {
        "name": "Web Video Text Track",
        "allow_subtitle": True,
        "scodecs": ["webvtt"],
        "max_subtitles": 1,
        "sstrict": True,
    },
    "avi": {
        "name": "Audio Video Interleave",
        "allow_video": True,
        "allow_audio": True,
        "vcodecs": ["mpeg4", "h264", "prores", "mjpeg", "mpeg2video", "rawvideo"],
        "acodecs": ["mp3", "aac", "vorbis", "mp2"],
        "disallow_v": ["hevc", "apng", "gif"],
    },
    "wmv": {
        "name": "Windows Media Video",
        "allow_video": True,
        "allow_audio": True,
        "vcodecs": ["msmpeg4v3", "h264", "mpeg4", "mpeg2video", "mjpeg", "rawvideo"],
        "acodecs": ["wmav2", "aac", "flac"],
        "disallow_v": ["prores", "hevc", "apng", "gif"],
    },
    "mkv": {
        "name": "Matroska",
        "allow_video": True,
        "allow_audio": True,
        "allow_subtitle": True,
        "allow_image": True,
        "vcodecs": [
            "h264",
            "hevc",
            "vp9",
            "vp8",
            "prores",
            "mpeg4",
            "mpeg2video",
            "msmpeg4v3",
            "mjpeg",
            "gif",
            "rawvideo",
        ],
        "acodecs": ["libvorbis", "vorbis", "opus", "flac", "aac", "mp2"],
        "disallow_v": ["apng"],
    },
    "mka": {
        "name": "Matroska Audio",
        "allow_audio": True,
        "acodecs": ["libvorbis", "vorbis", "opus", "flac", "aac", "mp2"],
    },
    "mov": {
        "name": "QuickTime / MOV",
        "allow_video": True,
        "allow_audio": True,
        "allow_subtitle": True,
        "vcodecs": [
            "h264",
            "hevc",
            "prores",
            "mpeg4",
            "mpeg2video",
            "msmpeg4v3",
            "mjpeg",
            "gif",
            "flv1",
            "rawvideo",
        ],
        "acodecs": ["aac", "mp3", "mp2", "vorbis"],
        "disallow_a": ["opus", "flac"],
        "disallow_v": ["apng", "vp9", "vp8"],
    },
    "swf": {
        "name": "ShockWave Flash / Small Web Format",
        "allow_video": True,
        "allow_audio": True,
        "vcodecs": ["flv1", "mjpeg"],
        "acodecs": ["mp3"],
        "vstrict": True,
        "astrict": True,
        "samplerate": [44100, 22050, 11025],
    },
    "not_in_here": {
        "allow_video": True,
        "allow_audio": True,
        "allow_subtitle": True,
    },
}


@dataclass
class Container:
    name: Optional[str] = None
    allow_video: bool = False
    allow_audio: bool = False
    allow_subtitle: bool = False
    allow_image: bool = False
    max_videos: Optional[int] = None
    max_audios: Optional[int] = None
    max_subtitles: Optional[int] = None
    vcodecs: Optional[List[str]] = None
    acodecs: Optional[List[str]] = None
    scodecs: Optional[List[str]] = None
    vstrict: bool = False
    astrict: bool = False
    sstrict: bool = False
    disallow_v: List[str] = field(default_factory=list)
    disallow_a: List[str] = field(default_factory=list)
    samplerate: Optional[List[int]] = None  # Any samplerate is allowed


def get_rules(key: str) -> Container:
    if key in containers:
        return Container(**containers[key])
    return Container(**containers["not_in_here"])



import numpy as np

# A NormalizeBuffer is a mono-channel circular buffer, used to normalize audio buffers.


class NormalizeBuffer:
    def __init__(self, length):
        self._data = np.zeros(length)
        self._offset = 0
        self.length = length

    def add(self, window):
        # Adds a window element-wise to the NormalizeBuffer.
        n = len(window)
        if n > self.length:
            raise ValueError("the window should be smaller than the NormalizeBuffer")

        # Compute the slice of data where the values will be added
        start = self._offset
        end = self._offset + n

        if end <= self.length:
            self._data[start:end] += window
        else:
            end -= self.length
            self._data[start:] += window[: self.length - start]
            self._data[:end] += window[self.length - start :]

    def remove(self, n):
        if n >= self.length:
            n = self.length
        if n == 0:
            return

        # Compute the slice of data to reset
        start = self._offset
        end = self._offset + n

        if end <= self.length:
            self._data[start:end] = 0
        else:
            end -= self.length
            self._data[start:] = 0
            self._data[:end] = 0

        self._offset += n
        self._offset %= self.length

    def to_array(self, start=0, end=None):
        if end is None:
            end = self.length

        start += self._offset
        end += self._offset

        if end <= self.length:
            return np.copy(self._data[start:end])

        end -= self.length
        if start < self.length:
            return np.concatenate((self._data[start:], self._data[:end]))

        start -= self.length
        return np.copy(self._data[start:end])



import sys


def desc_options(parser):
    parser.add_argument("--ffmpeg-location", help="Point to your custom ffmpeg file.")
    parser.add_required("input", nargs="*", help="Path to file(s)")
    return parser


def main(sys_args=sys.argv[1:]):
    from auto_editor.vanparse import ArgumentParser
    from auto_editor.ffwrapper import FFmpeg, FileInfo
    from auto_editor.utils.log import Log

    parser = desc_options(ArgumentParser("desc"))
    args = parser.parse_args(sys_args)

    ffmpeg = FFmpeg(args.ffmpeg_location, debug=False)

    print("")
    for input_file in args.input:
        inp = FileInfo(input_file, ffmpeg, Log())
        if "description" in inp.metadata:
            print(inp.metadata["description"], end="\n\n")
        else:
            print("No description.", end="\n\n")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

# Internal Libraries
import os
import sys
import tempfile

# Typing
from typing import List

# Included Libraries
import auto_editor
import auto_editor.vanparse as vanparse
import auto_editor.utils.func as usefulfunctions

from auto_editor.utils.progressbar import ProgressBar
from auto_editor.utils.log import Log, Timer
from auto_editor.ffwrapper import FFmpeg, FileInfo
from auto_editor.edit import edit_media


def main_options(parser):
    from auto_editor.objects import (
        TextObject,
        RectangleObject,
        EllipseObject,
        ImageObject,
    )
    from auto_editor.utils.types import (
        float_type,
        sample_rate_type,
        frame_type,
        range_type,
        speed_range_type,
        margin_type,
        color_type,
    )

    parser.add_text("Object Options")
    parser.add_argument(
        "--add-text",
        nargs="*",
        dataclass=TextObject,
        help="Add a text object to the timeline.",
    )
    parser.add_argument(
        "--add-rectangle",
        nargs="*",
        dataclass=RectangleObject,
        help="Add a rectangle object to the timeline.",
    )
    parser.add_argument(
        "--add-ellipse",
        nargs="*",
        dataclass=EllipseObject,
        help="Add an ellipse object to the timeline.",
    )
    parser.add_argument(
        "--add-image",
        nargs="*",
        dataclass=ImageObject,
        help="Add an image object onto the timeline.",
    )
    parser.add_text("URL Download Options")
    parser.add_argument(
        "--yt-dlp-location", default="yt-dlp", help="Set a custom path to yt-dlp."
    )
    parser.add_argument(
        "--download-format", help="Set the yt-dlp download format. (--format, -f)"
    )
    parser.add_argument(
        "--output-format", help="Set the yt-dlp output file template. (--output, -o)"
    )
    parser.add_argument(
        "--yt-dlp-extras", help="Add extra options for yt-dlp. Must be in quotes"
    )
    parser.add_text("Exporting as Media Options")
    parser.add_argument(
        "--video-codec",
        "-vcodec",
        "-c:v",
        default="auto",
        help="Set the video codec for the output media file.",
    )
    parser.add_argument(
        "--audio-codec",
        "-acodec",
        "-c:a",
        default="auto",
        help="Set the audio codec for the output media file.",
    )
    parser.add_argument(
        "--video-bitrate",
        "-b:v",
        default="10m",
        help="Set the number of bits per second for video.",
    )
    parser.add_argument(
        "--audio-bitrate",
        "-b:a",
        default="unset",
        help="Set the number of bits per second for audio.",
    )
    parser.add_argument(
        "--sample-rate",
        "-ar",
        type=sample_rate_type,
        help="Set the sample rate of the input and output videos.",
    )
    parser.add_argument(
        "--video-quality-scale",
        "-qscale:v",
        "-q:v",
        default="unset",
        help="Set a value to the ffmpeg option -qscale:v",
    )
    parser.add_argument(
        "--scale",
        type=float_type,
        default=1,
        help="Scale the output video by a certain factor.",
    )
    parser.add_argument(
        "--extras",
        help="Add extra options for ffmpeg for video rendering. Must be in quotes.",
    )
    parser.add_text("Miscellaneous Options")
    parser.add_argument(
        "--background",
        type=color_type,
        default="#000",
        help="Set the color of the background that is visible when the video is moved.",
    )
    parser.add_argument(
        "--combine-files",
        flag=True,
        help="Combine all input files into one before editing.",
    )
    parser.add_argument(
        "--progress",
        default="modern",
        choices=["modern", "classic", "ascii", "machine", "none"],
        help="Set what type of progress bar to use.",
    )
    parser.add_text("Manual Editing Options")
    parser.add_argument(
        "--cut-out",
        type=range_type,
        nargs="*",
        help="The range of media that will be removed completely, regardless of the "
        "value of silent speed.",
    )
    parser.add_argument(
        "--add-in",
        type=range_type,
        nargs="*",
        help="The range of media that will be added in, opposite of --cut-out",
    )
    parser.add_blank()
    parser.add_argument(
        "--mark-as-loud",
        type=range_type,
        nargs="*",
        help='The range that will be marked as "loud".',
    )
    parser.add_argument(
        "--mark-as-silent",
        type=range_type,
        nargs="*",
        help='The range that will be marked as "silent".',
    )
    parser.add_argument(
        "--set-speed-for-range",
        type=speed_range_type,
        nargs="*",
        help="Set an arbitrary speed for a given range.",
    )
    parser.add_text("Select Editing Source Options")
    parser.add_argument(
        "--edit-based-on",
        "--edit",
        default="audio",
        help="Decide which method to use when making edits.",
    )
    parser.add_argument(
        "--keep-tracks-seperate",
        flag=True,
        help="Don't combine audio tracks when exporting.",
    )
    parser.add_argument(
        "--export",
        default="default",
        choices=[
            "default",
            "premiere",
            "final-cut-pro",
            "shotcut",
            "json",
            "audio",
            "clip-sequence",
        ],
        help="Choose the export mode.",
    )
    parser.add_text("Utility Options")
    parser.add_argument(
        "--no-open", flag=True, help="Do not open the file after editing is done."
    )
    parser.add_argument(
        "--temp-dir",
        help="Set where the temporary directory is located.",
    )
    parser.add_argument(
        "--ffmpeg-location",
        help="Set a custom path to the ffmpeg location.",
    )
    parser.add_argument(
        "--my-ffmpeg",
        flag=True,
        help="Use the ffmpeg on your PATH instead of the one packaged.",
    )
    parser.add_text("Display Options")
    parser.add_argument(
        "--version", flag=True, help="Display the program's version and halt."
    )
    parser.add_argument(
        "--debug", flag=True, help="Show debugging messages and values."
    )
    parser.add_argument(
        "--show-ffmpeg-debug", flag=True, help="Show ffmpeg progress and output."
    )
    parser.add_argument("--quiet", "-q", flag=True, help="Display less output.")
    parser.add_argument(
        "--preview", flag=True, help="Show stats on how the input will be cut and halt."
    )
    parser.add_argument(
        "--timeline", flag=True, help="Show auto-editor JSON timeline file and halt."
    )
    parser.add_argument(
        "--api",
        default="0.2.0",
        help="Set what version of the JSON timeline to output.",
    )
    parser.add_text("Editing Options")
    parser.add_argument(
        "--silent-threshold",
        "-t",
        type=float_type,
        default=0.04,
        help="Set the volume that frames audio needs to surpass to be marked loud.",
    )
    parser.add_argument(
        "--frame-margin",
        "--margin",
        "-m",
        type=margin_type,
        default="6",
        help='Set how many "silent" frames on either side of the "loud" sections to include.',
    )
    parser.add_argument(
        "--silent-speed",
        "-s",
        type=float_type,
        default=99999,
        help='Set the speed that "silent" sections should be played at.',
    )
    parser.add_argument(
        "--video-speed",
        "--sounded-speed",
        "-v",
        type=float_type,
        default=1,
        help='Set the speed that "loud" sections should be played at.',
    )
    parser.add_argument(
        "--min-clip-length",
        "-mclip",
        type=frame_type,
        default=3,
        help="Set the minimum length a clip can be. If a clip is too short, cut it.",
    )
    parser.add_argument(
        "--min-cut-length",
        "-mcut",
        type=frame_type,
        default=6,
        help="Set the minimum length a cut can be. If a cut is too short, don't cut.",
    )
    parser.add_blank()
    parser.add_argument(
        "--output-file",
        "--output",
        "-o",
        nargs="*",
        help="Set the name(s) of the new output.",
    )
    parser.add_required(
        "input", nargs="*", help="File(s) or URL(s) that will be edited."
    )

    return parser


def main():
    parser = vanparse.ArgumentParser("Auto-Editor")

    subcommands = ["test", "info", "levels", "grep", "subdump", "desc"]

    if len(sys.argv) > 1 and sys.argv[1] in subcommands:
        obj = __import__(
            f"auto_editor.subcommands.{sys.argv[1]}", fromlist=["subcommands"]
        )
        obj.main(sys.argv[2:])
        sys.exit()
    else:
        parser = main_options(parser)

        # Preserve backwards compatibility

        sys_a = sys.argv[1:]

        def c(sys_a, options, new):
            for option in options:
                if option in sys_a:
                    pos = sys_a.index(option)
                    sys_a[pos : pos + 1] = new
            return sys_a

        sys_a = c(
            sys_a,
            ["--export_to_premiere", "--export-to-premiere", "-exp"],
            ["--export", "premiere"],
        )
        sys_a = c(
            sys_a,
            ["--export_to_final_cut_pro", "--export-to-final-cut-pro", "-exf"],
            ["--export", "final-cut-pro"],
        )
        sys_a = c(
            sys_a,
            ["--export_to_shotcut", "--export-to-shotcut", "-exs"],
            ["--export", "shotcut"],
        )
        sys_a = c(sys_a, ["--export_as_json", "--export-as-json"], ["--export", "json"])
        sys_a = c(
            sys_a,
            ["--export_as_clip_sequence", "--export-as-clip-sequence", "-excs"],
            ["--export", "clip-sequence"],
        )

        args = parser.parse_args(sys_a)

    timer = Timer(args.quiet)

    exporting_to_editor = args.export in ("premiere", "final-cut-pro", "shotcut")
    making_data_file = exporting_to_editor or args.export == "json"

    ffmpeg = FFmpeg(args.ffmpeg_location, args.my_ffmpeg, args.show_ffmpeg_debug)

    if args.debug and args.input == []:
        import platform as plat

        is64bit = "64-bit" if sys.maxsize > 2**32 else "32-bit"
        print(f"Python Version: {plat.python_version()} {is64bit}")
        print(f"Platform: {plat.system()} {plat.release()} {plat.machine().lower()}")
        print(f"FFmpeg Version: {ffmpeg.version}\nFFmpeg Path: {ffmpeg.path}")
        print(f"Auto-Editor Version: {auto_editor.version}")
        sys.exit()

    if args.version:
        print(f"Auto-Editor Version: {auto_editor.version}")
        sys.exit()

    if args.timeline:
        args.quiet = True

    if args.input == []:
        Log().error(
            "You need to give auto-editor an input file so it can do the work for you."
        )

    if args.temp_dir is None:
        TEMP = tempfile.mkdtemp()
    else:
        TEMP = args.temp_dir
        if os.path.isfile(TEMP):
            Log().error("Temp directory cannot be an already existing file.")
        if os.path.isdir(TEMP):
            if len(os.listdir(TEMP)) != 0:
                Log().error("Temp directory should be empty!")
        else:
            os.mkdir(TEMP)

    log = Log(args.debug, args.quiet, temp=TEMP)
    log.debug(f"Temp Directory: {TEMP}")

    log.conwrite("Starting")

    if args.preview or args.export not in ("audio", "default"):
        args.no_open = True

    if args.silent_speed <= 0 or args.silent_speed > 99999:
        args.silent_speed = 99999

    if args.video_speed <= 0 or args.video_speed > 99999:
        args.video_speed = 99999

    if args.output_file is None:
        args.output_file = []

    from auto_editor.validate_input import valid_input

    input_list = valid_input(args.input, ffmpeg, args, log)

    if args.combine_files:
        temp_file = (
            "combined.mp4"
            if exporting_to_editor
            else os.path.join(TEMP, "combined.mp4")
        )

        cmd = []
        for fileref in input_list:
            cmd.extend(["-i", fileref])
        cmd.extend(
            [
                "-filter_complex",
                f"[0:v]concat=n={len(input_list)}:v=1:a=1",
                "-codec:v",
                "h264",
                "-pix_fmt",
                "yuv420p",
                "-strict",
                "-2",
                temp_file,
            ]
        )

        ffmpeg.run(cmd)
        del cmd
        input_list = [temp_file]

    def main_loop(input_list: List[str], ffmpeg: FFmpeg, args, log: Log) -> None:
        num_cuts = 0

        progress = ProgressBar(args.progress)

        for i, input_path in enumerate(input_list):
            inp = FileInfo(input_path, ffmpeg, log)

            if len(input_list) > 1:
                log.conwrite(f"Working on {inp.basename}")

            cuts, output_path = edit_media(i, inp, ffmpeg, args, progress, TEMP, log)
            num_cuts += cuts

        if not args.preview and not args.timeline and not making_data_file:
            timer.stop()

        if not args.preview and not args.timeline and making_data_file:
            # Assume making each cut takes about 30 seconds.
            time_save = usefulfunctions.human_readable_time(num_cuts * 30)
            s = "s" if num_cuts != 1 else ""

            log.print(
                f"Auto-Editor made {num_cuts} cut{s}, which would have taken "
                f"about {time_save} if edited manually."
            )

        if not args.no_open and output_path is not None:
            usefulfunctions.open_with_system_default(output_path, log)

    try:
        main_loop(input_list, ffmpeg, args, log)
    except KeyboardInterrupt:
        log.error("Keyboard Interrupt")
    log.cleanup()


if __name__ == "__main__":
    main()



import sys
import json
import os.path

from typing import Optional

import av

av.logging.set_level(av.logging.PANIC)


def info_options(parser):
    parser.add_argument("--json", flag=True, help="Export info in JSON format.")
    parser.add_argument(
        "--include-vfr",
        "--has-vfr",
        flag=True,
        help="Display the number of Variable Frame Rate (VFR) frames.",
    )
    parser.add_argument(
        "--include-timebase",
        "--has-timebase",
        flag=True,
        help="Show what timebase the video streams have.",
    )
    parser.add_argument(
        "--ffmpeg-location", default=None, help="Point to your custom ffmpeg file."
    )
    parser.add_argument(
        "--my-ffmpeg",
        flag=True,
        help="Use the ffmpeg on your PATH instead of the one packaged.",
    )
    parser.add_required(
        "input", nargs="*", help="The path to a file you want inspected."
    )
    return parser


def main(sys_args=sys.argv[1:]):
    from auto_editor.utils.log import Log
    from auto_editor.utils.func import aspect_ratio
    from auto_editor.vanparse import ArgumentParser
    from auto_editor.ffwrapper import FFmpeg, FileInfo

    parser = info_options(ArgumentParser("info"))
    args = parser.parse_args(sys_args)

    ffmpeg = FFmpeg(args.ffmpeg_location, args.my_ffmpeg, False)

    file_info = {}

    for file in args.input:
        text = ""
        if os.path.exists(file):
            text += f"file: {file}\n"
        else:
            Log().error(f"Could not find file: {file}")

        inp = FileInfo(file, ffmpeg, Log())

        file_info[file] = {
            "video": [],
            "audio": [],
            "subtitle": [],
            "container": {},
        }

        if len(inp.videos) > 0:
            text += f" - video tracks: {len(inp.videos)}\n"

        for track, stream in enumerate(inp.videos):
            text += f"   - Track #{track}\n"
            text += f"     - codec: {stream.codec}\n"

            vid = {}
            vid["codec"] = stream.codec

            container = av.open(file, "r")

            pix_fmt = container.streams.video[track].pix_fmt
            text += f"     - pix_fmt: {pix_fmt}\n"
            vid["pix_fmt"] = pix_fmt

            if args.include_timebase:
                time_base = container.streams.video[track].time_base
                text += f"     - time_base: {time_base}\n"
                vid["time_base"] = time_base

            if stream.fps is not None:
                text += f"     - fps: {stream.fps}\n"
                vid["fps"] = float(stream.fps)

            w, h = stream.width, stream.height
            w_, h_ = aspect_ratio(w, h)
            text += f"     - resolution: {w}x{h} ({w_}:{h_})\n"

            vid["width"] = stream.width
            vid["height"] = stream.height
            vid["aspect_ratio"] = aspect_ratio(w, h)

            if stream.bitrate is not None:
                text += f"     - bitrate: {stream.bitrate}\n"
                vid["bitrate"] = stream.bitrate
            if stream.lang is not None:
                text += f"     - lang: {stream.lang}\n"
                vid["lang"] = stream.lang

            file_info[file]["video"].append(vid)

        if len(inp.audios) > 0:
            text += f" - audio tracks: {len(inp.audios)}\n"

        for track, stream in enumerate(inp.audios):
            aud = {}

            text += f"   - Track #{track}\n"
            text += f"     - codec: {stream.codec}\n"
            text += f"     - samplerate: {stream.samplerate}\n"

            aud["codec"] = stream.codec
            aud["samplerate"] = stream.samplerate

            if stream.bitrate is not None:
                text += f"     - bitrate: {stream.bitrate}\n"
                aud["bitrate"] = stream.bitrate

            if stream.lang is not None:
                text += f"     - lang: {stream.lang}\n"
                aud["lang"] = stream.lang

            file_info[file]["audio"].append(aud)

        if len(inp.subtitles) > 0:
            text += f" - subtitle tracks: {len(inp.subtitles)}\n"

        for track, stream in enumerate(inp.subtitles):
            sub = {}

            text += f"   - Track #{track}\n"
            text += f"     - codec: {stream.codec}\n"
            sub["codec"] = stream.codec
            if stream.lang is not None:
                text += f"     - lang: {stream.lang}\n"
                sub["lang"] = stream.lang

            file_info[file]["subtitle"].append(sub)

        if len(inp.videos) + len(inp.audios) + len(inp.subtitles) == 0:
            text += "Invalid media.\n"
            file_info[file] = {"media": "invalid"}
        else:
            text += " - container:\n"

            cont = file_info[file]["container"]

            if inp.duration is not None:
                text += f"   - duration: {inp.duration}\n"
                cont["duration"] = inp.duration
            if inp.bitrate is not None:
                text += f"   - bitrate: {inp.bitrate}\n"
                cont["bitrate"] = inp.bitrate

            if args.include_vfr:
                if not args.json:
                    print(text, end="")
                text = ""
                fps_mode = ffmpeg.pipe(
                    [
                        "-i",
                        file,
                        "-hide_banner",
                        "-vf",
                        "vfrdet",
                        "-an",
                        "-f",
                        "null",
                        "-",
                    ]
                )
                fps_mode = fps_mode.strip()

                if "VFR:" in fps_mode:
                    fps_mode = (fps_mode[fps_mode.index("VFR:") :]).strip()

                text += f"   - {fps_mode}\n"
                cont["fps_mode"] = fps_mode

        if not args.json:
            print(text)

    if args.json:
        json_object = json.dumps(file_info, indent=4)
        print(json_object)


if __name__ == "__main__":
    main()



from dataclasses import asdict, fields

from typing import List, Tuple, Dict, Union

from auto_editor.utils.log import Log
from auto_editor.ffwrapper import FileInfo


class Sheet:
    __slots__ = ("all", "sheet")

    def __init__(
        self, pool, inp: FileInfo, chunks: List[Tuple[int, int, float]], log: Log
    ) -> None:

        ending = chunks[:]
        if ending[-1][2] == 99999:
            ending.pop()

        end = 0
        if ending:
            end = ending[-1][1]

        _vars = {
            "width": inp.gwidth,
            "height": inp.gheight,
            "centerX": inp.gwidth // 2,
            "centerY": inp.gheight // 2,
            "start": 0,
            "end": end,
        }

        self.all = []
        self.sheet: Dict[int, List[int]] = {}

        def _values(val, _type, _vars: Dict[str, int]):
            if val is None:
                return None

            if _type is str:
                return str(val)  # Skip replacing variables with vals.

            for key, item in _vars.items():
                if val == key:
                    return _type(item)

            try:
                _type(val)
            except TypeError as e:
                log.error(str(e))
            except Exception:
                log.error(f"variable '{val}' is not defined.")

            return _type(val)

        for index, obj in enumerate(pool):

            dic_value = asdict(obj)
            dic_type = {}
            for field in fields(obj):
                dic_type[field.name] = field.type

            # Convert to the correct types
            for k, _type in dic_type.items():
                obj.__setattr__(k, _values(dic_value[k], _type, _vars))

            if obj.dur < 1:
                log.error(f"dur's value must be greater than 0. Was '{dur}'.")

            for frame in range(obj.start, obj.start + obj.dur, 1):
                if frame in self.sheet:
                    self.sheet[frame].append(index)
                else:
                    self.sheet[frame] = [index]

            self.all.append(obj)



import numpy as np

from .utils import windows, CBuffer, NormalizeBuffer

EPSILON = 0.0001


class AnalysisSynthesisTSM:
    def run(self, reader, writer, flush=True):
        finished = False
        while not (finished and reader.empty):
            self.read_from(reader)
            _, finished = self.write_to(writer)

        if flush:
            finished = False
            while not finished:
                _, finished = self.flush_to(writer)

            self.clear()

    def __init__(
        self,
        converter,
        channels,
        frame_length,
        analysis_hop,
        synthesis_hop,
        analysis_window,
        synthesis_window,
        delta_before=0,
        delta_after=0,
    ):
        self._converter = converter

        self._channels = channels
        self._frame_length = frame_length
        self._analysis_hop = analysis_hop
        self._synthesis_hop = synthesis_hop

        self._analysis_window = analysis_window
        self._synthesis_window = synthesis_window

        self._delta_before = delta_before
        self._delta_after = delta_after

        # When the analysis hop is larger than the frame length, some samples
        # from the input need to be skipped. self._skip_input_samples tracks
        # how many samples should be skipped before reading the analysis frame.
        self._skip_input_samples = 0

        # This attribute is used to start the output signal in the middle of a
        # frame, which should be the peek of the window function
        self._skip_output_samples = 0

        # Compute the normalize window
        self._normalize_window = windows.product(
            self._analysis_window, self._synthesis_window
        )

        if self._normalize_window is None:
            self._normalize_window = np.ones(self._frame_length)

        # Initialize the buffers
        delta = self._delta_before + self._delta_after
        self._in_buffer = CBuffer(self._channels, self._frame_length + delta)
        self._analysis_frame = np.empty((self._channels, self._frame_length + delta))
        self._out_buffer = CBuffer(self._channels, self._frame_length)
        self._normalize_buffer = NormalizeBuffer(self._frame_length)

        self.clear()

    def clear(self):
        # Clear the buffers
        self._in_buffer.remove(self._in_buffer.length)
        self._out_buffer.remove(self._out_buffer.length)
        self._out_buffer.right_pad(self._frame_length)
        self._normalize_buffer.remove(self._normalize_buffer.length)

        # Left pad the input with half a frame of zeros, and ignore that half
        # frame in the output. This makes the output signal start in the middle
        # of a frame, which should be the peak of the window function.
        self._in_buffer.write(
            np.zeros((self._channels, self._delta_before + self._frame_length // 2))
        )
        self._skip_output_samples = self._frame_length // 2

        self._converter.clear()

    def flush_to(self, writer):
        if self._in_buffer.remaining_length == 0:
            raise RuntimeError(
                "There is still data to process in the input buffer, flush_to method "
                "should only be called when write_to returns True."
            )

        n = self._out_buffer.write_to(writer)
        if self._out_buffer.ready == 0:
            # The output buffer is empty
            self.clear()
            return n, True

        return n, False

    def get_max_output_length(self, input_length):
        input_length -= self._skip_input_samples
        if input_length <= 0:
            return 0

        n_frames = input_length // self._analysis_hop + 1
        return n_frames * self._synthesis_hop

    def _process_frame(self):
        """Read an analysis frame from the input buffer, process it, and write
        the result to the output buffer."""
        # Generate the analysis frame and discard the input samples that will
        # not be needed anymore
        self._in_buffer.peek(self._analysis_frame)
        self._in_buffer.remove(self._analysis_hop)

        # Apply the analysis window
        windows.apply(self._analysis_frame, self._analysis_window)

        # Convert the analysis frame into a synthesis frame
        synthesis_frame = self._converter.convert_frame(self._analysis_frame)

        # Apply the synthesis window
        windows.apply(synthesis_frame, self._synthesis_window)

        # Overlap and add the synthesis frame in the output buffer
        self._out_buffer.add(synthesis_frame)

        # The overlap and add step changes the volume of the signal. The
        # normalize_buffer is used to keep track of "how much of the input
        # signal was added" to each part of the output buffer, allowing to
        # normalize it.
        self._normalize_buffer.add(self._normalize_window)

        # Normalize the samples that are ready to be written to the output
        normalize = self._normalize_buffer.to_array(end=self._synthesis_hop)
        normalize[normalize < EPSILON] = 1
        self._out_buffer.divide(normalize)
        self._out_buffer.set_ready(self._synthesis_hop)
        self._normalize_buffer.remove(self._synthesis_hop)

    def read_from(self, reader):
        n = reader.skip(self._skip_input_samples)
        self._skip_input_samples -= n
        if self._skip_input_samples > 0:
            return n

        n += self._in_buffer.read_from(reader)

        if (
            self._in_buffer.remaining_length == 0
            and self._out_buffer.remaining_length >= self._synthesis_hop
        ):
            # The input buffer has enough data to process, and there is enough
            # space in the output buffer to store the output
            self._process_frame()

            # Skip output samples if necessary
            skipped = self._out_buffer.remove(self._skip_output_samples)
            self._out_buffer.right_pad(skipped)
            self._skip_output_samples -= skipped

            # Set the number of input samples to be skipped
            self._skip_input_samples = self._analysis_hop - self._frame_length
            if self._skip_input_samples < 0:
                self._skip_input_samples = 0

        return n

    def set_speed(self, speed):
        self._analysis_hop = int(self._synthesis_hop * speed)
        self._converter.set_analysis_hop(self._analysis_hop)

    def write_to(self, writer):
        n = self._out_buffer.write_to(writer)
        self._out_buffer.right_pad(n)

        if self._in_buffer.remaining_length > 0 and self._out_buffer.ready == 0:
            # There is not enough data to process in the input buffer, and the
            # output buffer is empty
            return n, True

        return n, False



import os
from math import ceil
from dataclasses import dataclass, asdict, fields

from typing import List, Tuple, Union, Callable

import numpy as np

from auto_editor.wavfile import read
from auto_editor.utils.log import Log
from auto_editor.ffwrapper import FileInfo
from auto_editor.utils.progressbar import ProgressBar


def get_media_duration(path: str, fps: float, temp: str, log: Log) -> int:

    audio_path = os.path.join(temp, "0.wav")

    if os.path.isfile(audio_path):
        sample_rate, audio_samples = read(audio_path)
        sample_rate_per_frame = sample_rate / fps
        return ceil(audio_samples.shape[0] / sample_rate_per_frame)

    import av

    container = av.open(path)

    if len(container.streams.video) < 1:
        log.error("Could not get media duration")

    video = av.open(path, "r").streams.video[0]
    return int(float(video.duration * video.time_base) * fps)


def get_audio_list(
    stream: Union[int, str],
    threshold: float,
    fps: float,
    progress: ProgressBar,
    temp: str,
    log: Log,
) -> np.ndarray:

    from auto_editor.analyze.audio import audio_detection

    path = os.path.join(temp, f"{stream}.wav")

    if os.path.isfile(path):
        sample_rate, audio_samples = read(path)
    else:
        log.error(f"Audio stream '{stream}' does not exist.")

    audio_list = audio_detection(audio_samples, sample_rate, fps, progress)

    del audio_samples

    return np.fromiter((x > threshold for x in audio_list), dtype=np.bool_)


def operand_combine(a: np.ndarray, b: np.ndarray, call: Callable) -> np.ndarray:
    if len(a) > len(b):
        b = np.resize(b, len(a))
    if len(b) > len(a):
        a = np.resize(a, len(b))

    return call(a, b)


def get_stream_data(
    method: str,
    attrs,
    args,
    inp: FileInfo,
    progress: ProgressBar,
    temp: str,
    log: Log,
) -> np.ndarray:

    if method == "none":
        return np.ones(
            (get_media_duration(inp.path, inp.gfps, temp, log)), dtype=np.bool_
        )
    if method == "all":
        return np.zeros(
            (get_media_duration(inp.path, inp.gfps, temp, log)), dtype=np.bool_
        )
    if method == "audio":
        if attrs.stream == "all":
            total_list = None
            for i in range(len(inp.audios)):
                audio_list = get_audio_list(
                    i, attrs.threshold, inp.gfps, progress, temp, log
                )
                if total_list is None:
                    total_list = audio_list
                else:
                    total_list = operand_combine(total_list, audio_list, np.logical_or)

            if total_list is None:
                log.error("Input has no audio streams.")
            return total_list
        else:
            return get_audio_list(
                attrs.stream, attrs.threshold, inp.gfps, progress, temp, log
            )
    if method == "motion":
        from auto_editor.analyze.motion import motion_detection

        if len(inp.videos) == 0:
            log.error("Video stream '0' does not exist.")

        motion_list = motion_detection(inp, progress, attrs.width, attrs.blur)
        return np.fromiter((x >= attrs.threshold for x in motion_list), dtype=np.bool_)

    if method == "pixeldiff":
        from auto_editor.analyze.pixeldiff import pixel_difference

        if len(inp.videos) == 0:
            log.error("Video stream '0' does not exist.")

        pixel_list = pixel_difference(inp, progress)
        return np.fromiter((x >= attrs.threshold for x in pixel_list), dtype=np.bool_)


def get_attributes(attrs_str, dataclass, log: Log):
    from auto_editor.vanparse import parse_dataclass

    attrs = parse_dataclass(attrs_str, dataclass)

    dic_value = asdict(attrs)
    dic_type = {}
    for field in fields(attrs):
        dic_type[field.name] = field.type

    # Convert to the correct types
    for k, _type in dic_type.items():
        try:
            attrs.__setattr__(k, _type(dic_value[k]))
        except (ValueError, TypeError) as e:
            log.error(str(e))

    return attrs


def get_has_loud(
    method_str: str, inp: FileInfo, progress: ProgressBar, temp: str, log: Log, args
) -> np.ndarray:

    from auto_editor.utils.types import float_type

    def stream_type(val: str):
        if val == "all":
            return val
        return int(val)

    @dataclass
    class Audio:
        stream: stream_type = 0
        threshold: float_type = args.silent_threshold

    @dataclass
    class Motion:
        threshold: float_type = 0.02
        blur: int = 9
        width: int = 400

    @dataclass
    class Pixeldiff:
        threshold: int = 1

    KEYWORD_SEP = " "
    METHOD_ATTRS_SEP = ":"

    result_array = None
    operand = None

    logic_funcs = {
        "and": np.logical_and,
        "or": np.logical_or,
        "xor": np.logical_xor,
    }

    method_str = method_str.replace("_", " ")  # Allow old style `--edit` to work

    for method in method_str.split(KEYWORD_SEP):

        if method == "":  # Skip whitespace
            continue

        if METHOD_ATTRS_SEP in method:
            method, attrs_str = method.split(METHOD_ATTRS_SEP)
        else:
            attrs_str = ""

        if method == "audio":
            attrs = get_attributes(attrs_str, Audio, log)
        elif method == "motion":
            attrs = get_attributes(attrs_str, Motion, log)
        elif method == "pixeldiff":
            attrs = get_attributes(attrs_str, Pixeldiff, log)
        else:
            attrs = None

        if method in ("audio", "motion", "pixeldiff", "none", "all"):

            if result_array is not None and operand is None:
                log.error("Logic operator must be between two editing methods.")

            stream_data = get_stream_data(method, attrs, args, inp, progress, temp, log)

            if operand == "not":
                result_array = np.logical_not(stream_data)
                operand = None
            elif result_array is None:
                result_array = stream_data
            elif operand in ("and", "or", "xor"):
                result_array = operand_combine(
                    result_array, stream_data, logic_funcs[operand]
                )
                operand = None

        elif method in ("and", "or", "xor"):
            if operand is not None:
                log.error("Invalid Editing Syntax.")
            if result_array is None:
                log.error(f"'{method}' operand needs two arguments.")
            operand = method
        elif method == "not":
            if operand is not None:
                log.error("Invalid Editing Syntax.")
            operand = method
        else:
            log.error(f"Unknown method/operator: '{method}'")

    if operand is not None:
        log.error(f"Dangling operand: '{operand}'")

    return result_array


def get_chunks(
    inp: FileInfo, args, progress: ProgressBar, temp: str, log: Log
) -> List[Tuple[int, int, float]]:
    from auto_editor.cutting import (
        to_speed_list,
        set_range,
        chunkify,
        apply_mark_as,
        apply_margin,
        seconds_to_frames,
        cook,
    )

    start_margin, end_margin = args.frame_margin

    fps = inp.gfps

    start_margin = seconds_to_frames(start_margin, fps)
    end_margin = seconds_to_frames(end_margin, fps)
    min_clip = seconds_to_frames(args.min_clip_length, fps)
    min_cut = seconds_to_frames(args.min_cut_length, fps)

    has_loud = get_has_loud(args.edit_based_on, inp, progress, temp, log, args)
    has_loud_length = len(has_loud)

    has_loud = apply_mark_as(has_loud, has_loud_length, fps, args, log)
    has_loud = cook(has_loud, min_clip, min_cut)
    has_loud = apply_margin(has_loud, has_loud_length, start_margin, end_margin)

    # Remove small clips/cuts created by applying other rules.
    has_loud = cook(has_loud, min_clip, min_cut)

    speed_list = to_speed_list(has_loud, args.video_speed, args.silent_speed)

    if args.cut_out != []:
        speed_list = set_range(speed_list, args.cut_out, fps, 99999, log)

    if args.add_in != []:
        speed_list = set_range(speed_list, args.add_in, fps, args.video_speed, log)

    if args.set_speed_for_range != []:
        for item in args.set_speed_for_range:
            speed_list = set_range(speed_list, [item[1:]], fps, item[0], log)

    return chunkify(speed_list, has_loud_length)



# Internal Libraries
from datetime import timedelta

# Typing
from typing import List, Tuple, Union

# Included Libraries
from auto_editor.utils.func import get_new_length
from auto_editor.utils.log import Log
from auto_editor.ffwrapper import FileInfo


def display_length(secs: Union[int, float]) -> str:
    if secs < 0:
        return "-" + str(timedelta(seconds=round(abs(secs))))
    return str(timedelta(seconds=round(secs)))


def time_frame(title: str, frames: Union[int, float], fps: float) -> None:
    in_sec = round(frames / fps, 1)
    minutes = timedelta(seconds=round(in_sec))
    print(f"{title}: {in_sec} secs ({minutes})")


def preview(inp: FileInfo, chunks: List[Tuple[int, int, float]], log: Log) -> None:
    log.conwrite("")

    old_length = chunks[-1][1] / inp.gfps
    new_length = get_new_length(chunks, inp.gfps)

    diff = new_length - old_length

    print(
        f"\nlength:\n - change: ({display_length(old_length)}) 100% -> "
        f"({display_length(new_length)}) {round((new_length / old_length) * 100, 2)}%\n "
        f"- diff: ({display_length(diff)}) {round((diff / old_length) * 100, 2)}%"
    )

    clips = 0
    cuts = 0
    cut_lens = []
    clip_lens = []
    for chunk in chunks:
        if chunk[2] != 99999:
            clips += 1
            leng = (chunk[1] - chunk[0]) / chunk[2]
            clip_lens.append(leng)
        else:
            cuts += 1
            leng = chunk[1] - chunk[0]
            cut_lens.append(leng)

    print(f"clips: {clips}")
    if len(clip_lens) < 2:
        time_frame(" - clip length", sum(clip_lens), inp.gfps)
    else:
        time_frame(" - smallest", min(clip_lens), inp.gfps)
        time_frame(" - largest", max(clip_lens), inp.gfps)
        time_frame(" - average", sum(clip_lens) / len(clip_lens), inp.gfps)

    print(f"cuts: {cuts}")
    if len(cut_lens) < 2:
        time_frame(" - cut length", sum(cut_lens), inp.gfps)
    else:
        time_frame(" - smallest", min(cut_lens), inp.gfps)
        time_frame(" - largest", max(cut_lens), inp.gfps)
        time_frame(" - average", sum(cut_lens) / len(cut_lens), inp.gfps)
    print("")

    log.debug(f"Chunks: {chunks}")



import os.path
from os.path import abspath
from shutil import move
from urllib.parse import quote
from platform import system

from typing import List, Tuple

from .utils import indent, safe_mkdir

from auto_editor.ffwrapper import FileInfo

"""
Premiere Pro uses the Final Cut Pro 7 XML Interchange Format

See docs here:
https://developer.apple.com/library/archive/documentation/AppleApplications/Reference/FinalCutPro_XML/Elements/Elements.html

Also, Premiere itself will happily output subtlety incorrect XML files that don't
come back the way they started.
"""

PIXEL_ASPECT_RATIO = "square"
COLORDEPTH = "24"
ANAMORPHIC = "FALSE"
DEPTH = "16"


def get_samplerate(inp: FileInfo) -> str:
    if len(inp.audios) > 0 and inp.audios[0].samplerate is not None:
        return inp.audios[0].samplerate
    return "48000"


def fix_url(path: str) -> str:
    if system() == "Windows":
        return "file://localhost/" + quote(abspath(path)).replace("%5C", "/")
    return "file://localhost" + abspath(path)


def speedup(speed: float) -> str:
    return indent(
        6,
        "<filter>",
        "\t<effect>",
        "\t\t<name>Time Remap</name>",
        "\t\t<effectid>timeremap</effectid>",
        "\t\t<effectcategory>motion</effectcategory>",
        "\t\t<effecttype>motion</effecttype>",
        "\t\t<mediatype>video</mediatype>",
        '\t\t<parameter authoringApp="PremierePro">',
        "\t\t\t<parameterid>variablespeed</parameterid>",
        "\t\t\t<name>variablespeed</name>",
        "\t\t\t<valuemin>0</valuemin>",
        "\t\t\t<valuemax>1</valuemax>",
        "\t\t\t<value>0</value>",
        "\t\t</parameter>",
        '\t\t<parameter authoringApp="PremierePro">',
        "\t\t\t<parameterid>speed</parameterid>",
        "\t\t\t<name>speed</name>",
        "\t\t\t<valuemin>-100000</valuemin>",
        "\t\t\t<valuemax>100000</valuemax>",
        f"\t\t\t<value>{speed}</value>",
        "\t\t</parameter>",
        '\t\t<parameter authoringApp="PremierePro">',
        "\t\t\t<parameterid>reverse</parameterid>",
        "\t\t\t<name>reverse</name>",
        "\t\t\t<value>FALSE</value>",
        "\t\t</parameter>",
        '\t\t<parameter authoringApp="PremierePro">',
        "\t\t\t<parameterid>frameblending</parameterid>",
        "\t\t\t<name>frameblending</name>",
        "\t\t\t<value>FALSE</value>",
        "\t\t</parameter>",
        "\t</effect>",
        "</filter>",
    )


def premiere_xml(
    inp: FileInfo,
    temp: str,
    output: str,
    chunks: List[Tuple[int, int, float]],
) -> None:

    audio_file = len(inp.videos) == 0 and len(inp.audios) == 1

    # This is not at all how timebase works in actual media but that's how it works here.
    fps = inp.gfps
    timebase = int(fps)

    if fps == 23.98 or fps == 23.97602397 or fps == 23.976:
        timebase = 24
        ntsc = "TRUE"
    elif fps == 29.97 or fps == 29.97002997:
        timebase = 30
        ntsc = "TRUE"
    elif fps == 59.94 or fps == 59.94005994:
        timebase = 60
        ntsc = "TRUE"
    else:
        ntsc = "FALSE"

    duration = chunks[-1][1]

    clips = []
    for chunk in chunks:
        if chunk[2] != 99999:
            clips.append(chunk)

    pathurls = [fix_url(inp.path)]

    tracks = len(inp.audios)

    if tracks > 1:
        name_without_extension = inp.basename[: inp.basename.rfind(".")]

        fold = safe_mkdir(os.path.join(inp.dirname, f"{name_without_extension}_tracks"))

        for i in range(1, tracks):
            newtrack = os.path.join(fold, f"{i}.wav")
            move(os.path.join(temp, f"{i}.wav"), newtrack)
            pathurls.append(fix_url(newtrack))

    width, height = inp.gwidth, inp.gheight

    group_name = f"Auto-Editor {'Audio' if audio_file else 'Video'} Group"

    with open(output, "w", encoding="utf-8") as outfile:
        outfile.write('<?xml version="1.0" encoding="UTF-8"?>\n<!DOCTYPE xmeml>\n')
        outfile.write('<xmeml version="4">\n')
        outfile.write("\t<sequence>\n")
        outfile.write(f"\t\t<name>{group_name}</name>\n")
        outfile.write(f"\t\t<duration>{duration}</duration>\n")
        outfile.write("\t\t<rate>\n")
        outfile.write(f"\t\t\t<timebase>{timebase}</timebase>\n")
        outfile.write(f"\t\t\t<ntsc>{ntsc}</ntsc>\n")
        outfile.write("\t\t</rate>\n")
        outfile.write("\t\t<media>\n")
        outfile.write(
            indent(
                3,
                "<video>",
                "\t<format>",
                "\t\t<samplecharacteristics>",
            )
        )

        if len(inp.videos) > 0:
            outfile.write(
                indent(
                    3,
                    "\t\t\t<rate>",
                    f"\t\t\t\t<timebase>{timebase}</timebase>",
                    f"\t\t\t\t<ntsc>{ntsc}</ntsc>",
                    "\t\t\t</rate>",
                )
            )

        outfile.write(
            indent(
                3,
                f"\t\t\t<width>{width}</width>",
                f"\t\t\t<height>{height}</height>",
                f"\t\t\t<pixelaspectratio>{PIXEL_ASPECT_RATIO}</pixelaspectratio>",
            )
        )

        if len(inp.videos) > 0:
            outfile.write(
                indent(
                    3,
                    "\t\t\t<fielddominance>none</fielddominance>",
                    f"\t\t\t<colordepth>{COLORDEPTH}</colordepth>",
                )
            )

        outfile.write(
            indent(
                3,
                "\t\t</samplecharacteristics>",
                "\t</format>",
                "</video>" if len(inp.videos) == 0 else "\t<track>",
            )
        )

        if len(inp.videos) > 0:
            # Handle video clips

            total = 0.0
            for j, clip in enumerate(clips):

                clip_duration = (clip[1] - clip[0] + 1) / clip[2]

                _start = int(total)
                _end = int(total) + int(clip_duration)
                _in = int(clip[0] / clip[2])
                _out = int(clip[1] / clip[2])

                total += clip_duration

                outfile.write(
                    indent(
                        5,
                        f'<clipitem id="clipitem-{j+1}">',
                        "\t<masterclipid>masterclip-2</masterclipid>",
                        f"\t<name>{inp.basename}</name>",
                        f"\t<start>{_start}</start>",
                        f"\t<end>{_end}</end>",
                        f"\t<in>{_in}</in>",
                        f"\t<out>{_out}</out>",
                    )
                )

                if j == 0:
                    outfile.write(
                        indent(
                            6,
                            '<file id="file-1">',
                            f"\t<name>{inp.basename}</name>",
                            f"\t<pathurl>{pathurls[0]}</pathurl>",
                            "\t<rate>",
                            f"\t\t<timebase>{timebase}</timebase>",
                            f"\t\t<ntsc>{ntsc}</ntsc>",
                            "\t</rate>",
                            f"\t<duration>{duration}</duration>",
                            "\t<media>",
                            "\t\t<video>",
                            "\t\t\t<samplecharacteristics>",
                            "\t\t\t\t<rate>",
                            f"\t\t\t\t\t<timebase>{timebase}</timebase>",
                            f"\t\t\t\t\t<ntsc>{ntsc}</ntsc>",
                            "\t\t\t\t</rate>",
                            f"\t\t\t\t<width>{width}</width>",
                            f"\t\t\t\t<height>{height}</height>",
                            f"\t\t\t\t<anamorphic>{ANAMORPHIC}</anamorphic>",
                            f"\t\t\t\t<pixelaspectratio>{PIXEL_ASPECT_RATIO}</pixelaspectratio>",
                            "\t\t\t\t<fielddominance>none</fielddominance>",
                            "\t\t\t</samplecharacteristics>",
                            "\t\t</video>",
                            "\t\t<audio>",
                            "\t\t\t<samplecharacteristics>",
                            f"\t\t\t\t<depth>{DEPTH}</depth>",
                            f"\t\t\t\t<samplerate>{inp.gsamplerate}</samplerate>",
                            "\t\t\t</samplecharacteristics>",
                            "\t\t\t<channelcount>2</channelcount>",
                            "\t\t</audio>",
                            "\t</media>",
                            "</file>",
                        )
                    )
                else:
                    outfile.write('\t\t\t\t\t\t<file id="file-1"/>\n')

                if clip[2] != 1:
                    outfile.write(speedup(clip[2] * 100))

                # Linking for video blocks
                for i in range(max(3, tracks + 1)):
                    outfile.write("\t\t\t\t\t\t<link>\n")
                    outfile.write(
                        f"\t\t\t\t\t\t\t<linkclipref>clipitem-{(i*(len(clips)))+j+1}</linkclipref>\n"
                    )
                    if i == 0:
                        outfile.write("\t\t\t\t\t\t\t<mediatype>video</mediatype>\n")
                    else:
                        outfile.write("\t\t\t\t\t\t\t<mediatype>audio</mediatype>\n")
                    if i == 2:
                        outfile.write("\t\t\t\t\t\t\t<trackindex>2</trackindex>\n")
                    else:
                        outfile.write("\t\t\t\t\t\t\t<trackindex>1</trackindex>\n")
                    outfile.write(f"\t\t\t\t\t\t\t<clipindex>{j+1}</clipindex>\n")
                    if i > 0:
                        outfile.write("\t\t\t\t\t\t\t<groupindex>1</groupindex>\n")
                    outfile.write("\t\t\t\t\t\t</link>\n")

                outfile.write("\t\t\t\t\t</clipitem>\n")
            outfile.write(indent(3, "\t</track>", "</video>"))

        # Audio Clips
        outfile.write(
            indent(
                3,
                "<audio>",
                "\t<numOutputChannels>2</numOutputChannels>",
                "\t<format>",
                "\t\t<samplecharacteristics>",
                f"\t\t\t<depth>{DEPTH}</depth>",
                f"\t\t\t<samplerate>{inp.gsamplerate}</samplerate>",
                "\t\t</samplecharacteristics>",
                "\t</format>",
            )
        )

        for t in range(tracks):
            outfile.write(
                '\t\t\t\t<track currentExplodedTrackIndex="0" premiereTrackType="Stereo">\n'
            )
            total = 0
            for j, clip in enumerate(clips):

                clip_duration = (clip[1] - clip[0] + 1) / clip[2]

                _start = int(total)
                _end = int(total) + int(clip_duration)
                _in = int(clip[0] / clip[2])
                _out = int(clip[1] / clip[2])

                total += clip_duration

                if audio_file:
                    clip_item_num = j + 1
                    master_id = "1"
                else:
                    clip_item_num = len(clips) + 1 + j + (t * len(clips))
                    master_id = "2"

                outfile.write(
                    indent(
                        5,
                        f'<clipitem id="clipitem-{clip_item_num}" premiereChannelType="stereo">',
                        f"\t<masterclipid>masterclip-{master_id}</masterclipid>",
                        f"\t<name>{inp.basename}</name>",
                        f"\t<start>{_start}</start>",
                        f"\t<end>{_end}</end>",
                        f"\t<in>{_in}</in>",
                        f"\t<out>{_out}</out>",
                    )
                )

                if j == 0 and (audio_file or t > 0):
                    outfile.write(
                        indent(
                            6,
                            f'<file id="file-{t+1}">',
                            f"\t<name>{inp.basename}</name>",
                            f"\t<pathurl>{pathurls[t]}</pathurl>",
                            "\t<rate>",
                            f"\t\t<timebase>{timebase}</timebase>",
                            f"\t\t<ntsc>{ntsc}</ntsc>",
                            "\t</rate>",
                            "\t<media>",
                            "\t\t<audio>",
                            "\t\t\t<samplecharacteristics>",
                            f"\t\t\t\t<depth>{DEPTH}</depth>",
                            f"\t\t\t\t<samplerate>{inp.gsamplerate}</samplerate>",
                            "\t\t\t</samplecharacteristics>",
                            "\t\t\t<channelcount>2</channelcount>",
                            "\t\t</audio>",
                            "\t</media>",
                            "</file>",
                        )
                    )
                else:
                    outfile.write(f'\t\t\t\t\t\t<file id="file-{t+1}"/>\n')

                outfile.write(
                    indent(
                        6,
                        "<sourcetrack>",
                        "\t<mediatype>audio</mediatype>",
                        "\t<trackindex>1</trackindex>",
                        "</sourcetrack>",
                        "<labels>",
                        "\t<label2>Iris</label2>",
                        "</labels>",
                    )
                )

                if clip[2] != 1:
                    outfile.write(speedup(clip[2] * 100))

                outfile.write("\t\t\t\t\t</clipitem>\n")
            if not audio_file:
                outfile.write("\t\t\t\t\t<outputchannelindex>1</outputchannelindex>\n")
            outfile.write("\t\t\t\t</track>\n")

        outfile.write("\t\t\t</audio>\n")
        outfile.write("\t\t</media>\n")
        outfile.write("\t</sequence>\n")
        outfile.write("</xmeml>\n")



from typing import List, Tuple, Union
from auto_editor.utils.log import Log

"""
To prevent duplicate code being pasted between scripts, common functions should be
put here. No function should modify or create video/audio files on its own.
"""


def get_stdout(cmd: List[str]) -> str:
    from subprocess import Popen, PIPE, STDOUT

    stdout, _ = Popen(cmd, stdout=PIPE, stderr=STDOUT).communicate()
    return stdout.decode("utf-8", "replace")


def clean_list(x: List[str], rm_chars: str) -> List[str]:
    new_list = []
    for item in x:
        for char in rm_chars:
            item = item.replace(char, "")
        new_list.append(item)
    return new_list


def aspect_ratio(width: int, height: int) -> Union[Tuple[int, int], Tuple[None, None]]:
    if height == 0:
        return None, None

    def gcd(a: int, b: int) -> int:
        while b:
            a, b = b, a % b
        return a

    c = gcd(width, height)
    return width // c, height // c


def get_new_length(chunks: List[Tuple[int, int, float]], fps: float) -> float:
    time_in_frames = 0.0
    for chunk in chunks:
        leng = chunk[1] - chunk[0]
        if chunk[2] != 99999:
            time_in_frames += leng * (1 / chunk[2])
    return time_in_frames / fps


def human_readable_time(time_in_secs: Union[int, float]) -> str:
    units = "seconds"
    if time_in_secs >= 3600:
        time_in_secs = round(time_in_secs / 3600, 1)
        if time_in_secs % 1 == 0:
            time_in_secs = round(time_in_secs)
        units = "hours"
    if time_in_secs >= 60:
        time_in_secs = round(time_in_secs / 60, 1)
        if time_in_secs >= 10 or time_in_secs % 1 == 0:
            time_in_secs = round(time_in_secs)
        units = "minutes"
    return f"{time_in_secs} {units}"


def open_with_system_default(path: str, log: Log) -> None:
    from subprocess import run

    try:  # should work on Windows
        from os import startfile  # type: ignore

        startfile(path)
    except OSError:
        pass
    except ImportError:
        try:  # should work on MacOS and most Linux versions
            run(["open", path])
        except Exception:
            try:  # should work on WSL2
                run(["cmd.exe", "/C", "start", path])
            except Exception:
                try:  # should work on various other Linux distros
                    run(["xdg-open", path])
                except Exception:
                    log.warning("Could not open output file.")


def fnone(val: object) -> bool:
    return val == "none" or val == "unset" or val is None


def append_filename(path: str, val: str) -> str:
    from os.path import splitext

    root, ext = splitext(path)
    return root + val + ext



# Internal Libraries
import os.path
import subprocess

# Typing
from typing import List, Tuple

# Included Libraries
from auto_editor.utils.log import Log
from auto_editor.ffwrapper import FFmpeg, FileInfo
from auto_editor.utils.progressbar import ProgressBar

from auto_editor.utils.encoder import encoders
from auto_editor.utils.video import get_vcodec, video_quality

# From: github.com/PyAV-Org/PyAV/blob/main/av/video/frame.pyx
allowed_pix_fmt = {
    "yuv420p",
    "yuvj420p",
    "rgb48be",
    "rgb48le",
    "rgb64be",
    "rgb64le",
    "rgb24",
    "bgr24",
    "argb",
    "rgba",
    "abgr",
    "bgra",
    "gray",
    "gray8",
    "gray16be",
    "gray16le",
    "rgb8",
    "bgr8",
    "pal8",
}


def pix_fmt_allowed(pix_fmt: str) -> bool:
    return pix_fmt in allowed_pix_fmt


def apply_anchor(
    x: int, y: int, width: int, height: int, anchor: str
) -> Tuple[int, int]:
    if anchor == "ce":
        x = int((x * 2 - width) / 2)
        y = int((y * 2 - height) / 2)
    if anchor == "tr":
        x -= width
    if anchor == "bl":
        y -= height
    if anchor == "br":
        x -= width
        y -= height
    # Pillow uses 'tl' by default
    return x, y


def one_pos_two_pos(
    x: int, y: int, width: int, height: int, anchor: str
) -> Tuple[int, int, int, int]:
    """Convert: x, y, width, height -> x1, y1, x2, y2"""

    if anchor == "ce":
        x1 = x - int(width / 2)
        x2 = x + int(width / 2)
        y1 = y - int(height / 2)
        y2 = y + int(height / 2)

        return x1, y1, x2, y2

    if anchor in ("tr", "br"):
        x1 = x - width
        x2 = x
    else:
        x1 = x
        x2 = x + width

    if anchor in ("tl", "tr"):
        y1 = y
        y2 = y + height
    else:
        y1 = y
        y2 = y - height

    return x1, y1, x2, y2


def render_av(
    ffmpeg: FFmpeg,
    track: int,
    inp: FileInfo,
    args,
    chunks: List[Tuple[int, int, float]],
    progress: ProgressBar,
    effects,
    rules,
    temp: str,
    log: Log,
) -> Tuple[str, bool]:
    try:
        import av

        av.logging.set_level(av.logging.PANIC)
    except ImportError:
        log.import_error("av")
    try:
        from PIL import Image, ImageDraw, ImageFont, ImageChops
    except ImportError:
        log.import_error("Pillow")

    def set_static_assets(all_objects, log: Log):
        # Save reloading the same thing over and over.

        new_objects = []

        for obj in all_objects:
            if obj._type == "text":
                try:
                    obj.font = ImageFont.truetype(obj.font, obj.size)
                except OSError:
                    if obj.font == "default":
                        obj.font = ImageFont.load_default()
                    else:
                        log.error(f"Font '{obj.font}' not found.")

            if obj._type == "image":
                source = Image.open(obj.src)
                source = source.convert("RGBA")
                source = source.rotate(obj.rotate, expand=True)
                source = ImageChops.multiply(
                    source,
                    Image.new(
                        "RGBA", source.size, (255, 255, 255, int(obj.opacity * 255))
                    ),
                )
                obj.src = source

            new_objects.append(obj)

        return new_objects

    def render_objects(sheet, all_objects, index: int, frame, pix_fmt: str):
        img = frame.to_image().convert("RGBA")

        for item in sheet[index]:
            obj = all_objects[item]

            obj_img = Image.new("RGBA", img.size, (255, 255, 255, 0))
            draw = ImageDraw.Draw(obj_img)

            if obj._type == "text":
                text_w, text_h = draw.textsize(obj.content, font=obj.font)
                pos = apply_anchor(obj.x, obj.y, text_w, text_h, "ce")
                draw.text(
                    pos,
                    obj.content,
                    font=obj.font,
                    fill=obj.fill,
                    align=obj.align,
                    stroke_width=obj.stroke,
                    stroke_fill=obj.strokecolor,
                )

            if obj._type == "rectangle":
                draw.rectangle(
                    one_pos_two_pos(obj.x, obj.y, obj.width, obj.height, obj.anchor),
                    fill=obj.fill,
                    width=obj.stroke,
                    outline=obj.strokecolor,
                )

            if obj._type == "ellipse":
                draw.ellipse(
                    one_pos_two_pos(obj.x, obj.y, obj.width, obj.height, obj.anchor),
                    fill=obj.fill,
                    width=obj.stroke,
                    outline=obj.strokecolor,
                )

            if obj._type == "image":
                img_w, img_h = obj.src.size
                pos = apply_anchor(obj.x, obj.y, img_w, img_h, obj.anchor)
                obj_img.paste(obj.src, pos)

            img = Image.alpha_composite(img, obj_img)

        return frame.from_image(img).reformat(format=pix_fmt)

    if chunks[-1][2] == 99999:
        chunks.pop()

    total_frames = chunks[-1][1]
    progress.start(total_frames, "Creating new video")

    fps = inp.videos[track].fps

    container = av.open(inp.path, "r")
    pix_fmt = container.streams.video[track].pix_fmt

    target_pix_fmt = pix_fmt

    if not pix_fmt_allowed(pix_fmt):
        target_pix_fmt = "yuv420p"

    my_codec = get_vcodec(args, inp, rules)

    apply_video_later = True

    if my_codec in encoders:
        apply_video_later = encoders[my_codec]["pix_fmt"].isdisjoint(allowed_pix_fmt)

    if args.scale != 1:
        apply_video_later = False

    log.debug(f"apply video quality settings now: {not apply_video_later}")

    video_stream = container.streams.video[track]
    video_stream.thread_type = "AUTO"

    width = video_stream.width
    height = video_stream.height

    effects.all = set_static_assets(effects.all, log)

    spedup = os.path.join(temp, f"spedup{track}.mp4")

    cmd = [
        "-hide_banner",
        "-y",
        "-f",
        "rawvideo",
        "-c:v",
        "rawvideo",
        "-pix_fmt",
        target_pix_fmt,
        "-s",
        f"{width}*{height}",
        "-framerate",
        f"{fps}",
        "-i",
        "-",
        "-pix_fmt",
        target_pix_fmt,
    ]

    if apply_video_later:
        cmd.extend(["-c:v", "mpeg4", "-qscale:v", "1"])
    else:
        cmd = video_quality(cmd, args, inp, rules)

    cmd.append(spedup)

    process2 = ffmpeg.Popen(
        cmd, stdin=subprocess.PIPE, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
    )

    input_equavalent = 0.0
    output_equavalent = 0
    chunk = chunks.pop(0)

    tou = int(video_stream.time_base.denominator / fps)
    log.debug(f"Tou {tou}")

    seek = 10
    seek_frame = None
    frames_saved = 0

    SEEK_COST = int(
        fps * 5
    )  # Keyframes are usually spread out every 5 seconds or less.
    SEEK_RETRY = SEEK_COST // 2

    try:
        for frame in container.decode(video_stream):
            index = int(frame.time * fps)

            if frame.key_frame:
                log.debug(f"Keyframe {index} {frame.pts}")

            if seek_frame is not None:
                log.debug(f"Skipped {index - seek_frame} frames")
                frames_saved += index - seek_frame
                seek_frame = None

            if index > chunk[1]:
                if chunks:
                    chunk = chunks.pop(0)
                else:
                    break

            if chunk[2] == 99999:
                if chunk[1] - index > SEEK_COST and index > seek:
                    seek = index + SEEK_RETRY

                    seek_frame = index
                    container.seek(chunk[1] * tou, stream=video_stream)
            else:
                input_equavalent += 1 / chunk[2]

            while input_equavalent > output_equavalent:
                if index in effects.sheet:
                    frame = render_objects(
                        effects.sheet, effects.all, index, frame, target_pix_fmt
                    )
                elif pix_fmt != target_pix_fmt:
                    frame = frame.reformat(format=target_pix_fmt)

                in_bytes = frame.to_ndarray().tobytes()
                process2.stdin.write(in_bytes)
                output_equavalent += 1

            if index % 3 == 0:
                progress.tick(index)

        progress.end()
        process2.stdin.close()
        process2.wait()
    except (OSError, BrokenPipeError):
        progress.end()
        ffmpeg.run_check_errors(cmd, log, True)
        log.error("FFmpeg Error!")

    log.debug(f"Total frames saved seeking: {frames_saved}")

    # Unfortunately, scaling has to be a concrete step.
    if args.scale != 1:
        sped_input = os.path.join(temp, f"spedup{track}.mp4")
        spedup = os.path.join(temp, f"scale{track}.mp4")

        cmd = [
            "-i",
            sped_input,
            "-vf",
            f"scale=iw*{args.scale}:ih*{args.scale}",
            spedup,
        ]

        check_errors = ffmpeg.pipe(cmd)
        if "Error" in check_errors or "failed" in check_errors:
            if "-allow_sw 1" in check_errors:
                cmd.insert(-1, "-allow_sw")
                cmd.insert(-1, "1")
            # Run again to show errors even if it might not work.
            ffmpeg.run(cmd)

    return spedup, apply_video_later



# Internal Libraries
import os
import sys
import shutil
import logging
import platform
import subprocess
from time import perf_counter

# External Libraries
import av
import numpy as np

av.logging.set_level(av.logging.PANIC)

# Typing
from typing import List, Tuple, Callable, NoReturn, Optional

# Included Libraries
from auto_editor.vanparse import ArgumentParser


def test_options(parser):
    parser.add_argument("--only", "-n", nargs="*")
    parser.add_argument(
        "--help",
        "-h",
        flag=True,
        help="Print info about the program or an option and exit.",
    )
    parser.add_required(
        "category",
        nargs=1,
        choices=["cli", "sub", "api", "unit", "all"],
        help="Set what category of tests to run.",
    )
    return parser


def pipe_to_console(cmd: List[str]) -> Tuple[int, str, str]:
    process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = process.communicate()
    return process.returncode, stdout.decode("utf-8"), stderr.decode("utf-8")


def cleanup(the_dir: str) -> None:
    for item in os.listdir(the_dir):
        item = os.path.join(the_dir, item)
        if (
            "_ALTERED" in item
            or item.endswith(".xml")
            or item.endswith(".fcpxml")
            or item.endswith(".mlt")
        ):
            os.remove(item)
        if item.endswith("_tracks"):
            shutil.rmtree(item)


def clean_all() -> None:
    cleanup("resources")
    cleanup(os.getcwd())


def get_runner() -> List[str]:
    if platform.system() == "Windows":
        return ["py", "-m", "auto_editor"]
    return ["python3", "-m", "auto_editor"]


def run_program(cmd: List[str]) -> None:
    no_open = "." in cmd[0]
    cmd = get_runner() + cmd

    if no_open:
        cmd += ["--no_open"]

    returncode, stdout, stderr = pipe_to_console(cmd)
    if returncode > 0:
        raise Exception(f"{stdout}\n{stderr}\n")


def check_for_error(cmd: List[str], match=None) -> None:
    returncode, stdout, stderr = pipe_to_console(get_runner() + cmd)
    if returncode > 0:
        if "Error!" in stderr:
            if match is not None and match not in stderr:
                raise Exception(f'Could\'t find "{match}"')
        else:
            raise Exception(f"Program crashed.\n{stdout}\n{stderr}")
    else:
        raise Exception("Program should not respond with a code 0.")


def make_np_list(in_file: str, compare_file: str, the_speed: float) -> None:
    from auto_editor.wavfile import read
    from auto_editor.render.tsm import phasevocoder, ArrReader, ArrWriter

    _, sped_chunk = read(in_file)
    channels = 2

    reader = ArrReader(sped_chunk)
    writer = ArrWriter(np.zeros((0, 2), dtype=np.int16))

    phasevocoder(channels, speed=the_speed).run(reader, writer)

    spedup_audio = writer.output
    del writer
    del reader

    loaded = np.load(compare_file)

    if not np.array_equal(spedup_audio, loaded["a"]):
        if spedup_audio.shape == loaded["a"].shape:
            print(f"Both shapes ({spedup_audio.shape}) are same")
        else:
            print(spedup_audio.shape)
            print(loaded["a"].shape)

        result = np.subtract(spedup_audio, loaded["a"])

        print(f"result non-zero: {np.count_nonzero(result)}")
        print(f"len of spedup_audio: {len(spedup_audio)}")

        print(
            np.count_nonzero(result) / spedup_audio.shape[0],
            "difference between arrays",
        )

        raise Exception(f"file {compare_file} doesn't match array.")

    # np.savez_compressed(out_file, a=spedup_audio)


class Tester:
    def __init__(self, args):
        self.passed_tests = 0
        self.failed_tests = 0
        self.args = args

    def run_test(self, func: Callable, cleanup=None, allow_fail=False) -> None:
        if self.args.only != [] and func.__name__ not in self.args.only:
            return

        start = perf_counter()
        try:
            func()
            end = perf_counter() - start
        except KeyboardInterrupt:
            print(f"Testing Interrupted by User.")
            clean_all()
            sys.exit(1)
        except Exception as e:
            self.failed_tests += 1
            print(f"Test '{func.__name__}' failed.\n{e}")
            if not allow_fail:
                logging.error("", exc_info=True)
                clean_all()
                sys.exit(1)
        else:
            self.passed_tests += 1
            print(f"Test '{func.__name__}' passed: {round(end, 2)} secs")
            if cleanup is not None:
                cleanup()

    def end(self) -> NoReturn:
        print(f"{self.passed_tests}/{self.passed_tests + self.failed_tests}")
        clean_all()
        sys.exit(0)


def main(sys_args: Optional[List[str]] = None):
    parser = test_options(ArgumentParser("test"))

    if sys_args is None:
        sys_args = sys.argv[1:]

    args = parser.parse_args(sys_args)

    ### Tests ###

    ## API Tests ##

    def read_api_0_1():
        check_for_error(
            ["resources/json/0.1-non-zero-start.json"],
            "Error! First chunk must start with 0",
        )
        check_for_error(
            ["resources/json/0.1-disjoint.json"], "Error! Chunk disjointed at"
        )

    def help_tests():
        """check the help option, its short, and help on options and groups."""
        run_program(["--help"])
        run_program(["-h"])
        run_program(["--frame_margin", "--help"])
        run_program(["--frame_margin", "-h"])
        run_program(["--help", "--help"])
        run_program(["-h", "--help"])
        run_program(["--help", "-h"])
        run_program(["-h", "--help"])

    def version_test():
        """Test version flags and debug by itself."""
        run_program(["--version"])
        run_program(["-v"])
        run_program(["-V"])
        run_program(["--debug"])

    def parser_test():
        check_for_error(["example.mp4", "--video-speed"], "needs argument")

    def subtitle_tests():
        from auto_editor.render.subtitle import SubtitleParser

        test = SubtitleParser()
        test.contents = [
            [0, 10, "A"],
            [10, 20, "B"],
            [20, 30, "C"],
            [30, 40, "D"],
            [40, 50, "E"],
            [50, 60, "F"],
        ]
        chunks = [
            (0, 10, 1),
            (10, 20, 99999),
            (20, 30, 1),
            (30, 40, 99999),
            (40, 50, 1),
            (50, 60, 99999),
        ]
        test.edit(chunks)

        if test.contents != [[0, 10, "A"], [10, 20, "C"], [20, 30, "E"]]:
            raise ValueError("Incorrect subtitle results.")

    def tsm_1a5_test():
        make_np_list(
            "resources/wav/example-cut-s16le.wav",
            "resources/data/example_1.5_speed.npz",
            1.5,
        )

    def tsm_0a5_test():
        make_np_list(
            "resources/wav/example-cut-s16le.wav",
            "resources/data/example_0.5_speed.npz",
            0.5,
        )

    def tsm_2a0_test():
        make_np_list(
            "resources/wav/example-cut-s16le.wav",
            "resources/data/example_2.0_speed.npz",
            2,
        )

    def info():
        run_program(["info", "example.mp4"])
        run_program(["info", "resources/only-video/man-on-green-screen.mp4"])
        run_program(["info", "resources/multi-track.mov"])
        run_program(["info", "resources/new-commentary.mp3"])
        run_program(["info", "resources/testsrc.mkv"])

    def levels():
        run_program(["levels", "resources/multi-track.mov"])
        run_program(["levels", "resources/new-commentary.mp3"])

    def subdump():
        run_program(["subdump", "resources/subtitle.mp4"])

    def grep():
        run_program(["grep", "boop", "resources/subtitle.mp4"])

    def desc():
        run_program(["desc", "example.mp4"])

    def example_tests():
        run_program(["example.mp4", "--video_codec", "uncompressed"])
        with av.open("example_ALTERED.mp4") as cn:
            video = cn.streams.video[0]
            assert video.average_rate == 30
            assert video.width == 1280
            assert video.height == 720
            assert video.codec.name == "mpeg4"
            assert cn.streams.audio[0].codec.name == "aac"
            assert cn.streams.audio[0].rate == 48000

        run_program(["example.mp4"])
        with av.open("example_ALTERED.mp4") as cn:
            video = cn.streams.video[0]
            assert video.average_rate == 30
            assert video.width == 1280
            assert video.height == 720
            assert video.codec.name == "h264"
            assert cn.streams.audio[0].codec.name == "aac"
            assert cn.streams.audio[0].rate == 48000
            assert video.language == "eng"
            assert cn.streams.audio[0].language == "eng"

    # PR #260
    def high_speed_test():
        run_program(["example.mp4", "--video-speed", "99998"])

    # Issue #200
    def url_test():
        run_program(["https://github.com/WyattBlue/auto-editor/raw/master/example.mp4"])

    # Issue #184
    def unit_tests():
        """
        Make sure all units are working appropriately. That includes:

         - Seconds units: s, sec, secs, second, seconds
         - Frame units:   f, frame, frames
         - Sample units:  Hz, kHz
         - Percent:       %
        """

        run_program(
            ["example.mp4", "--mark_as_loud", "20s,22sec", "25secs,26.5seconds"]
        )
        run_program(["example.mp4", "--sample_rate", "44100"])
        run_program(["example.mp4", "--sample_rate", "44100 Hz"])
        run_program(["example.mp4", "--sample_rate", "44.1 kHz"])
        run_program(["example.mp4", "--silent_threshold", "4%"])

    def backwards_range_test():
        """
        Cut out the last 5 seconds of a media file by using negative number in the
        range.
        """
        run_program(["example.mp4", "--edit", "none", "--cut_out", "-5secs,end"])
        run_program(["example.mp4", "--edit", "all", "--add_in", "-5secs,end"])

    def cut_out_test():
        run_program(
            [
                "example.mp4",
                "--edit",
                "none",
                "--video_speed",
                "2",
                "--silent_speed",
                "3",
                "--cut_out",
                "2secs,10secs",
            ]
        )
        run_program(
            [
                "example.mp4",
                "--edit",
                "all",
                "--video_speed",
                "2",
                "--add_in",
                "2secs,10secs",
            ]
        )

    def gif_test():
        """
        Feed auto-editor a gif file and make sure it can spit out a correctly formated
        gif. No editing is requested.
        """
        run_program(["resources/only-video/man-on-green-screen.gif", "--edit", "none"])
        with av.open("resources/only-video/man-on-green-screen_ALTERED.gif") as cn:
            assert cn.streams.video[0].codec.name == "gif"

    def margin_tests():
        run_program(["example.mp4", "-m", "3"])
        run_program(["example.mp4", "--margin", "3"])
        run_program(["example.mp4", "-m", "0.3sec"])
        run_program(["example.mp4", "-m", "6f,-3secs"])
        run_program(["example.mp4", "-m", "3,5 frames"])
        run_program(["example.mp4", "-m", "0.4 seconds"])

    def input_extension():
        """Input file must have an extension. Throw error if none is given."""

        shutil.copy("example.mp4", "example")
        check_for_error(["example", "--no_open"], "must have an extension.")
        os.remove("example")

    def output_extension():
        # Add input extension to output name if no output extension is given.
        run_program(["example.mp4", "-o", "out"])
        with av.open("out.mp4") as cn:
            assert cn.streams.video[0].codec.name == "h264"

        os.remove("out.mp4")

        run_program(["resources/testsrc.mkv", "-o", "out"])
        with av.open("out.mkv") as cn:
            assert cn.streams.video[0].codec.name == "h264"

        os.remove("out.mkv")

    def progress_ops_test():
        run_program(["example.mp4", "--progress", "machine"])
        run_program(["example.mp4", "--progress", "none"])
        run_program(["example.mp4", "--progress", "ascii"])

    def silent_threshold():
        run_program(["resources/new-commentary.mp3", "--silent_threshold", "0.1"])

    def track_tests():
        run_program(["resources/multi-track.mov", "--keep_tracks_seperate"])

    def json_tests():
        run_program(["example.mp4", "--export_as_json"])
        run_program(["example.json"])

    def scale_tests():
        run_program(["example.mp4", "--scale", "1.5"])
        with av.open("example_ALTERED.mp4") as cn:
            assert cn.streams.video[0].average_rate == 30
            assert cn.streams.video[0].width == 1920
            assert cn.streams.video[0].height == 1080
            assert cn.streams.audio[0].rate == 48000

        run_program(["example.mp4", "--scale", "0.2"])
        with av.open("example_ALTERED.mp4") as cn:
            assert cn.streams.video[0].average_rate == 30
            assert cn.streams.video[0].width == 256
            assert cn.streams.video[0].height == 144
            assert cn.streams.audio[0].rate == 48000

    def various_errors_test():
        check_for_error(
            ["example.mp4", "--add_rectangle", "0,60", "--cut_out", "60,end"]
        )

    def effect_tests():
        """Test rendering video objects"""
        run_program(
            [
                "resources/testsrc.mp4",
                "--mark_as_loud",
                "start,end",
                "--add_rectangle",
                "0,30,0,200,100,300,fill=#43FA56,stroke=10",
            ]
        )
        os.remove("resources/testsrc_ALTERED.mp4")

    def render_text():
        run_program(["example.mp4", "--add-text", "start,end,This is my text"])

    def check_font_error():
        check_for_error(
            ["example.mp4", "--add-text", "start,end,text,0,0,30,notafont"], "not found"
        )

    def export_tests():
        for test_name in (
            "aac.m4a",
            "alac.m4a",
            "wav/pcm-f32le.wav",
            "wav/pcm-s32le.wav",
            "multi-track.mov",
            "subtitle.mp4",
            "testsrc.mkv",
        ):

            test_file = f"resources/{test_name}"
            run_program([test_file])
            run_program([test_file, "--edit", "none"])
            run_program([test_file, "-exp"])
            run_program([test_file, "-exf"])
            run_program([test_file, "-exs"])
            run_program([test_file, "--export_as_clip_sequence"])
            run_program([test_file, "--preview"])
            cleanup("resources")

    def codec_tests():
        run_program(["example.mp4", "--video_codec", "h264"])
        run_program(["example.mp4", "--audio_codec", "ac3"])

    def combine_tests():
        run_program(["example.mp4", "--mark_as_silent", "0,171", "-o", "hmm.mp4"])
        run_program(["example.mp4", "hmm.mp4", "--combine_files", "--debug"])
        os.remove("hmm.mp4")

    def image_test():
        run_program(["resources/embedded-image/h264-png.mp4"])
        with av.open("resources/embedded-image/h264-png_ALTERED.mp4", "r") as cn:
            assert cn.streams.video[0].codec.name == "h264"
            assert cn.streams.video[1].codec.name == "png"

        run_program(["resources/embedded-image/h264-mjpeg.mp4"])
        with av.open("resources/embedded-image/h264-mjpeg_ALTERED.mp4", "r") as cn:
            assert cn.streams.video[0].codec.name == "h264"
            assert cn.streams.video[1].codec.name == "mjpeg"

        run_program(["resources/embedded-image/h264-png.mkv"])
        with av.open("resources/embedded-image/h264-png_ALTERED.mkv", "r") as cn:
            assert cn.streams.video[0].codec.name == "h264"
            assert cn.streams.video[1].codec.name == "png"

        run_program(["resources/embedded-image/h264-mjpeg.mkv"])
        with av.open("resources/embedded-image/h264-mjpeg_ALTERED.mkv", "r") as cn:
            assert cn.streams.video[0].codec.name == "h264"
            assert cn.streams.video[1].codec.name == "mjpeg"

    def motion_tests():
        run_program(
            [
                "resources/only-video/man-on-green-screen.mp4",
                "--edit",
                "motion",
                "--debug",
                "--frame_margin",
                "0",
                "-mcut",
                "0",
                "-mclip",
                "0",
            ]
        )
        run_program(
            [
                "resources/only-video/man-on-green-screen.mp4",
                "--edit",
                "motion:threshold=0",
            ]
        )

    def edit_positive_tests():
        run_program(["resources/multi-track.mov", "--edit", "audio:stream=all"])
        run_program(["resources/multi-track.mov", "--edit", "not audio:stream=all"])
        run_program(
            [
                "resources/multi-track.mov",
                "--edit",
                "not audio:threshold=4% or audio:stream=1",
            ]
        )
        # run_program(['resources/multi-track.mov', '--edit', 'not audio:threshold=4% or not audio:stream=1'])

    def edit_negative_tests():
        check_for_error(
            ["resources/wav/example-cut-s16le.wav", "--edit", "motion"],
            "Video stream '0' does not exist",
        )
        check_for_error(
            ["resources/only-video/man-on-green-screen.gif", "--edit", "audio"],
            "Audio stream '0' does not exist",
        )
        check_for_error(
            ["example.mp4", "--edit", "not"], "Error! Dangling operand: 'not'"
        )
        check_for_error(
            ["example.mp4", "--edit", "audio and"], "Error! Dangling operand: 'and'"
        )
        check_for_error(
            ["example.mp4", "--edit", "and"],
            "Error! 'and' operand needs two arguments.",
        )
        check_for_error(
            ["example.mp4", "--edit", "and audio"],
            "Error! 'and' operand needs two arguments.",
        )
        check_for_error(
            ["example.mp4", "--edit", "or audio"],
            "Error! 'or' operand needs two arguments.",
        )
        check_for_error(
            ["example.mp4", "--edit", "audio four audio"],
            "Error! Unknown method/operator: 'four'",
        )
        check_for_error(
            ["example.mp4", "--edit", "audio audio"],
            "Logic operator must be between two editing methods",
        )

    ### Runners ###

    tester = Tester(args)

    if args.category in ("unit", "all"):
        tester.run_test(subtitle_tests)
        tester.run_test(tsm_1a5_test)
        tester.run_test(tsm_0a5_test, allow_fail=True)
        tester.run_test(tsm_2a0_test)

    if args.category in ("api", "all"):
        tester.run_test(read_api_0_1)

    if args.category in ("sub", "all"):
        tester.run_test(info)
        tester.run_test(levels)
        tester.run_test(subdump)
        tester.run_test(grep)
        tester.run_test(desc)

    if args.category in ("cli", "all"):
        tester.run_test(help_tests)
        tester.run_test(version_test)
        tester.run_test(parser_test)
        tester.run_test(example_tests)
        tester.run_test(high_speed_test)
        tester.run_test(url_test)
        tester.run_test(unit_tests)
        tester.run_test(backwards_range_test)
        tester.run_test(cut_out_test)
        tester.run_test(image_test)
        tester.run_test(gif_test, cleanup=clean_all)
        tester.run_test(margin_tests)
        tester.run_test(input_extension)
        tester.run_test(output_extension)
        tester.run_test(progress_ops_test)
        tester.run_test(silent_threshold)
        tester.run_test(track_tests)
        tester.run_test(json_tests)
        tester.run_test(scale_tests)
        tester.run_test(various_errors_test)
        tester.run_test(effect_tests, cleanup=clean_all)
        tester.run_test(render_text)
        tester.run_test(check_font_error)
        tester.run_test(export_tests)
        tester.run_test(codec_tests)
        tester.run_test(combine_tests)
        tester.run_test(motion_tests)
        tester.run_test(edit_positive_tests)
        tester.run_test(edit_negative_tests)

    tester.end()


if __name__ == "__main__":
    main()



from typing import List, Tuple, NoReturn, Union

import numpy as np

from auto_editor.utils.log import Log
from auto_editor.utils.types import split_num_str


def remove_small(
    has_loud: np.ndarray, lim: int, replace: int, with_: int
) -> np.ndarray:
    startP = 0
    active = False
    for j, item in enumerate(has_loud):
        if item == replace:
            if not active:
                startP = j
                active = True
            # Special case for end.
            if j == len(has_loud) - 1:
                if j - startP < lim:
                    has_loud[startP : j + 1] = with_
        else:
            if active:
                if j - startP < lim:
                    has_loud[startP:j] = with_
                active = False
    return has_loud


def str_is_number(val: str) -> bool:
    return val.replace(".", "", 1).replace("-", "", 1).isdigit()


def str_starts_with_number(val: str) -> bool:
    if val.startswith("-"):
        val = val[1:]
    val = val.replace(".", "", 1)
    return val[0].isdigit()


def set_range(
    has_loud: np.ndarray, range_syntax: list, fps: float, with_: int, log: Log
) -> np.ndarray:
    def replace_variables_to_values(
        val: str, fps: float, log: Log
    ) -> Union[int, NoReturn]:
        if str_is_number(val):
            return int(val)
        if str_starts_with_number(val):
            try:
                value, unit = split_num_str(val)
            except TypeError as e:
                log.error(str(e))

            if unit in ("", "f", "frame", "frames"):
                if isinstance(value, float):
                    log.error("float type cannot be used with frame unit")
                return int(value)
            if unit in ("s", "sec", "secs", "second", "seconds"):
                return round(value * fps)
            log.error(f"Unknown unit: {unit}")

        if val == "start":
            return 0
        if val == "end":
            return len(has_loud)
        return log.error(f"variable '{val}' not available.")

    def var_val_to_frames(val: str, fps: float, log: Log) -> int:
        num = replace_variables_to_values(val, fps, log)
        if num < 0:
            num += len(has_loud)
        return num

    for item in range_syntax:
        pair = []
        for val in item:
            pair.append(var_val_to_frames(val, fps, log))
        has_loud[pair[0] : pair[1]] = with_
    return has_loud


def seconds_to_frames(value: Union[int, str], fps: float) -> int:
    if isinstance(value, str):
        return int(float(value) * fps)
    return value


def cook(has_loud: np.ndarray, min_clip: int, min_cut: int) -> np.ndarray:
    has_loud = remove_small(has_loud, min_clip, replace=1, with_=0)
    has_loud = remove_small(has_loud, min_cut, replace=0, with_=1)
    return has_loud


# Turn long silent/loud array to formatted chunk list.
# Example: [1, 1, 1, 2, 2] => [(0, 3, 1), (3, 5, 2)]
def chunkify(arr: np.ndarray, arr_length: int = None) -> List[Tuple[int, int, float]]:
    if arr_length is None:
        arr_length = len(arr)

    chunks = []
    start = 0
    for j in range(1, arr_length):
        if arr[j] != arr[j - 1]:
            chunks.append((start, j, arr[j - 1]))
            start = j
    chunks.append((start, arr_length, arr[j]))
    return chunks


def apply_margin(
    has_loud: np.ndarray, has_loud_length: int, start_m: int, end_m: int
) -> np.ndarray:

    # Find start and end indexes.
    start_index = []
    end_index = []
    for j in range(1, has_loud_length):
        if has_loud[j] != has_loud[j - 1]:
            if has_loud[j]:
                start_index.append(j)
            else:
                end_index.append(j)

    # Apply margin
    if start_m > 0:
        for i in start_index:
            has_loud[max(i - start_m, 0) : i] = True
    if start_m < 0:
        for i in start_index:
            has_loud[i : min(i - start_m, has_loud_length)] = False

    if end_m > 0:
        for i in end_index:
            has_loud[i : min(i + end_m, has_loud_length)] = True
    if end_m < 0:
        for i in end_index:
            has_loud[max(i + end_m, 0) : i] = False

    return has_loud


def apply_mark_as(
    has_loud: np.ndarray, has_loud_length: int, fps: float, args, log: Log
) -> np.ndarray:

    if args.mark_as_loud != []:
        has_loud = set_range(has_loud, args.mark_as_loud, fps, args.video_speed, log)

    if args.mark_as_silent != []:
        has_loud = set_range(has_loud, args.mark_as_silent, fps, args.silent_speed, log)
    return has_loud


def to_speed_list(
    has_loud: np.ndarray, video_speed: float, silent_speed: float
) -> np.ndarray:

    speed_list = has_loud.astype(float)

    # This code will break is speed is allowed to be 0
    speed_list[speed_list == 1] = video_speed
    speed_list[speed_list == 0] = silent_speed

    return speed_list


def merge(start_list: np.ndarray, end_list: np.ndarray) -> np.ndarray:
    merge = np.zeros((len(start_list)), dtype=np.bool_)

    startP = 0
    for item in start_list:
        if item == True:
            where_list = np.where(end_list[startP:])[0]
            if len(where_list) > 0:
                merge[startP : where_list[0]] = True
        startP += 1
    return merge



import wave

import numpy as np

from auto_editor.wavfile import read
from auto_editor.render.tsm import phasevocoder, ArrReader, ArrWriter
from auto_editor.utils.log import Log
from auto_editor.utils.progressbar import ProgressBar

from typing import List, Tuple


def make_new_audio(
    input_path: str,
    output_path: str,
    chunks: List[Tuple[int, int, float]],
    log: Log,
    fps: float,
    progress: ProgressBar,
) -> None:

    if len(chunks) == 1 and chunks[0][2] == 99999:
        log.error("Trying to create an empty file.")

    progress.start(len(chunks), "Creating new audio")

    samplerate, audio_samples = read(input_path)

    main_writer = wave.open(output_path, "wb")
    main_writer.setnchannels(2)
    main_writer.setframerate(samplerate)
    main_writer.setsampwidth(2)

    for c, chunk in enumerate(chunks):
        sample_start = int(chunk[0] / fps * samplerate)
        sample_end = int(sample_start + (samplerate / fps) * (chunk[1] - chunk[0]))

        the_speed = chunk[2]

        if the_speed == 1:
            main_writer.writeframes(audio_samples[sample_start:sample_end])
        elif the_speed != 99999:
            sped_chunk = audio_samples[sample_start:sample_end]

            reader = ArrReader(sped_chunk)
            writer = ArrWriter(np.zeros((0, 2), dtype=np.int16))

            phasevocoder(2, speed=the_speed).run(reader, writer)
            if writer.output.shape[0] != 0:
                main_writer.writeframes(writer.output)

        progress.tick(c)
    progress.end()



# Internal Libraries
import os.path

# Typing
from typing import List, Tuple, Optional

# Included Libraries
from .func import fnone
from .container import Container
from auto_editor.utils.log import Log
from auto_editor.ffwrapper import FFmpeg, FileInfo


def fset(cmd: List[str], option: str, value: str) -> List[str]:
    if fnone(value):
        return cmd
    return cmd + [option] + [value]


def get_vcodec(args, inp: FileInfo, rules: Container) -> str:
    vcodec = args.video_codec
    if vcodec == "auto":
        vcodec = inp.videos[0].codec

        if rules.vcodecs is not None:
            if rules.vstrict and vcodec not in rules.vcodecs:
                return rules.vcodecs[0]

            if vcodec in rules.disallow_v:
                return rules.vcodecs[0]

    if vcodec == "copy":
        return inp.videos[0].codec

    if vcodec == "uncompressed":
        return "mpeg4"
    return vcodec


def get_acodec(args, inp: FileInfo, rules: Container) -> str:
    acodec = args.audio_codec
    if acodec == "auto":
        acodec = inp.audios[0].codec

        if rules.acodecs is not None:  # Just in case, but shouldn't happen
            if rules.astrict and acodec not in rules.acodecs:
                # Input codec can't be used for output, so use a new safe codec.
                return rules.acodecs[0]

            if acodec in rules.disallow_a:
                return rules.acodecs[0]

    if acodec == "copy":
        return inp.audios[0].codec
    return acodec


def video_quality(cmd: List[str], args, inp: FileInfo, rules: Container) -> List[str]:
    cmd = fset(cmd, "-b:v", args.video_bitrate)

    qscale = args.video_quality_scale

    if args.video_codec == "uncompressed" and fnone(qscale):
        qscale = "1"

    vcodec = get_vcodec(args, inp, rules)

    cmd.extend(["-c:v", vcodec])

    cmd = fset(cmd, "-qscale:v", qscale)

    cmd.extend(["-movflags", "faststart"])
    return cmd


def mux_quality_media(
    ffmpeg: FFmpeg,
    video_output: List[Tuple[int, bool, str, bool]],
    rules: Container,
    write_file: str,
    container: str,
    args,
    inp: FileInfo,
    temp: str,
    log: Log,
) -> None:
    s_tracks = 0 if not rules.allow_subtitle else len(inp.subtitles)
    a_tracks = 0 if not rules.allow_audio else len(inp.audios)
    v_tracks = 0 if not rules.allow_video else len(video_output)

    cmd = ["-hide_banner", "-y", "-i", inp.path]

    # fmt: off
    for _, is_video, path, _, in video_output:
        if is_video or rules.allow_image:
            cmd.extend(["-i", path])
        else:
            v_tracks -= 1
    # fmt: on

    if a_tracks > 0:
        if args.keep_tracks_seperate and rules.max_audios is None:
            for t in range(a_tracks):
                cmd.extend(["-i", os.path.join(temp, f"new{t}.wav")])
        else:
            # Merge all the audio a_tracks into one.
            new_a_file = os.path.join(temp, "new_audio.wav")
            if a_tracks > 1:
                new_cmd = []
                for t in range(a_tracks):
                    new_cmd.extend(["-i", os.path.join(temp, f"new{t}.wav")])
                new_cmd.extend(
                    [
                        "-filter_complex",
                        f"amerge=inputs={a_tracks}",
                        "-ac",
                        "2",
                        new_a_file,
                    ]
                )
                ffmpeg.run(new_cmd)
                a_tracks = 1
            else:
                new_a_file = os.path.join(temp, "new0.wav")
            cmd.extend(["-i", new_a_file])

    if s_tracks > 0:
        for s, sub in enumerate(inp.subtitles):
            cmd.extend(["-i", os.path.join(temp, f"new{s}s.{sub.ext}")])

    total_streams = v_tracks + s_tracks + a_tracks

    for i in range(total_streams):
        cmd.extend(["-map", f"{i+1}:0"])

    cmd.extend(["-map_metadata", "0"])

    for track, is_video, path, apply_video in video_output:
        if is_video:
            if apply_video:
                cmd = video_quality(cmd, args, inp, rules)
            else:
                cmd.extend([f"-c:v:{track}", "copy"])
        elif rules.allow_image:
            ext = os.path.splitext(path)[1][1:]
            cmd.extend(
                [f"-c:v:{track}", ext, f"-disposition:v:{track}", "attached_pic"]
            )

    # Copy lang metadata
    streams = (
        (inp.videos, "v", v_tracks),
        (inp.audios, "a", a_tracks),
        (inp.subtitles, "s", s_tracks),
    )

    for stream, marker, max_streams in streams:
        for i, track in enumerate(stream):
            if i > max_streams:
                break
            if track.lang is not None:
                cmd.extend([f"-metadata:s:{marker}:{i}", f"language={track.lang}"])

    if s_tracks > 0:
        scodec = inp.subtitles[0].codec
        if inp.ext == f".{container}":
            cmd.extend(["-c:s", scodec])
        elif rules.scodecs is not None:
            if scodec not in rules.scodecs:
                scodec = rules.scodecs[0]
            cmd.extend(["-c:s", scodec])

    if a_tracks > 0:
        acodec = get_acodec(args, inp, rules)

        cmd = fset(cmd, "-c:a", acodec)
        cmd = fset(cmd, "-b:a", args.audio_bitrate)

        if fnone(args.sample_rate):
            if rules.samplerate is not None:
                cmd.extend(["-ar", str(rules.samplerate[0])])
        else:
            cmd.extend(["-ar", str(args.sample_rate)])

    if args.extras is not None:
        cmd.extend(args.extras.split(" "))
    cmd.extend(["-strict", "-2"])  # Allow experimental codecs.
    cmd.extend(
        ["-map", "0:t?", "-map", "0:d?"]
    )  # Add input attachments and data to output.
    cmd.append(write_file)
    ffmpeg.run_check_errors(cmd, log, path=write_file)



import numpy as np
import numpy.typing as npt
from math import ceil

from auto_editor.utils.progressbar import ProgressBar


def get_max_volume(s: np.ndarray) -> float:
    return max(float(np.max(s)), -float(np.min(s)))


def audio_detection(
    audio_samples: np.ndarray,
    sample_rate: int,
    fps: float,
    progress: ProgressBar,
) -> npt.NDArray[np.float_]:

    max_volume = get_max_volume(audio_samples)
    sample_count = audio_samples.shape[0]

    sample_rate_per_frame = sample_rate / fps
    audio_frame_count = ceil(sample_count / sample_rate_per_frame)

    progress.start(audio_frame_count, "Analyzing audio volume")

    threshold_list = np.zeros((audio_frame_count), dtype=np.float_)

    # Calculate when the audio is loud or silent.
    for i in range(audio_frame_count):

        if i % 500 == 0:
            progress.tick(i)

        start = int(i * sample_rate_per_frame)
        end = min(int((i + 1) * sample_rate_per_frame), sample_count)

        threshold_list[i] = get_max_volume(audio_samples[start:end]) / max_volume

    progress.end()
    return threshold_list



# Internal Libraries
import os

# Typing
from typing import List, Tuple, Optional

# Included Libraries
from auto_editor.sheet import Sheet
from auto_editor.utils.log import Log
from auto_editor.utils.func import fnone, append_filename
from auto_editor.utils.progressbar import ProgressBar
from auto_editor.method import get_chunks
from auto_editor.ffwrapper import FFmpeg, FileInfo


def set_output_name(path: str, inp_ext: str, export: str) -> str:
    root, ext = os.path.splitext(path)

    if export == "json":
        return root + ".json"
    if export == "final-cut-pro":
        return root + ".fcpxml"
    if export == "shotcut":
        return root + ".mlt"
    if export == "premiere":
        return root + ".xml"
    if export == "audio":
        return root + "_ALTERED.wav"
    if ext == "":
        return root + inp_ext

    return root + "_ALTERED" + ext


def edit_media(
    i: int,
    inp: FileInfo,
    ffmpeg: FFmpeg,
    args,
    progress: ProgressBar,
    temp: str,
    log: Log,
) -> Tuple[int, Optional[str]]:
    from auto_editor.utils.container import get_rules

    chunks = None
    if inp.ext == ".json":
        from auto_editor.formats.timeline import read_json_timeline

        args.background, input_path, chunks = read_json_timeline(inp.path, log)
        inp = FileInfo(input_path, ffmpeg, log)

    if i < len(args.output_file):
        output_path = args.output_file[i]

        # Add input extension if output doesn't have one.
        if os.path.splitext(output_path)[1] == "":
            output_path = set_output_name(output_path, inp.ext, args.export)
    else:
        output_path = set_output_name(inp.path, inp.ext, args.export)

    log.debug(f"{inp.path} -> {output_path}")

    output_container = os.path.splitext(output_path)[1].replace(".", "")

    # Check if export options make sense.
    rules = get_rules(output_container)
    codec_error = "'{}' codec is not supported in '{}' container."

    if not fnone(args.sample_rate):
        if rules.samplerate is not None and args.sample_rate not in rules.samplerate:
            log.error(
                f"'{output_container}' container only supports samplerates: {rules.samplerate}"
            )

    vcodec = args.video_codec
    if vcodec == "uncompressed":
        vcodec = "mpeg4"
    if vcodec == "copy":
        vcodec = inp.videos[0].codec

    if vcodec != "auto":
        if rules.vstrict and vcodec not in rules.vcodecs:
            log.error(codec_error.format(vcodec, output_container))

        if vcodec in rules.disallow_v:
            log.error(codec_error.format(vcodec, output_container))

    acodec = args.audio_codec
    if acodec == "copy":
        acodec = inp.audios[0].codec
        log.debug(f"Settings acodec to {acodec}")

    if acodec not in ("unset", "auto"):
        if rules.astrict and acodec not in rules.acodecs:
            log.error(codec_error.format(acodec, output_container))

        if acodec in rules.disallow_a:
            log.error(codec_error.format(acodec, output_container))

    if args.keep_tracks_seperate and rules.max_audios == 1:
        log.warning(
            f"'{output_container}' container doesn't support multiple audio tracks."
        )

    if not args.preview and not args.timeline:
        if os.path.isdir(output_path):
            log.error("Output path already has an existing directory!")

        if os.path.isfile(output_path) and inp.path != output_path:
            log.debug(f"Removing already existing file: {output_path}")
            os.remove(output_path)

    tracks = len(inp.audios)

    # Extract subtitles in their native format.
    if len(inp.subtitles) > 0:
        cmd = ["-i", inp.path, "-hide_banner"]
        for s, sub in enumerate(inp.subtitles):
            cmd.extend(["-map", f"0:s:{s}"])
        for s, sub in enumerate(inp.subtitles):
            cmd.extend([os.path.join(temp, f"{s}s.{sub.ext}")])
        ffmpeg.run(cmd)

    # Split audio tracks into: 0.wav, 1.wav, etc.
    log.conwrite("Extracting audio")

    cmd = ["-i", inp.path, "-hide_banner"]
    for t in range(tracks):
        cmd.extend(
            [
                "-map",
                f"0:a:{t}",
                "-ac",
                "2",
                "-rf64",
                "always",
                os.path.join(temp, f"{t}.wav"),
            ]
        )

    ffmpeg.run(cmd)
    del cmd

    if chunks is None:
        chunks = get_chunks(inp, args, progress, temp, log)

    if len(chunks) == 1 and chunks[0][2] == 99999:
        log.error("The entire media is cut!")

    def is_clip(chunk: Tuple[int, int, float]) -> bool:
        return chunk[2] != 99999

    def number_of_cuts(chunks: List[Tuple[int, int, float]]) -> int:
        return len(list(filter(is_clip, chunks)))

    num_cuts = number_of_cuts(chunks)

    pool = args.add_text + args.add_rectangle + args.add_ellipse + args.add_image

    obj_sheet = Sheet(pool, inp, chunks, log)

    if args.timeline:
        from auto_editor.formats.timeline import make_json_timeline

        make_json_timeline(
            args.api, inp.path, 0, obj_sheet, chunks, args.background, log
        )
        return num_cuts, None

    if args.preview:
        from auto_editor.preview import preview

        preview(inp, chunks, log)
        return num_cuts, None

    if args.export == "json":
        from auto_editor.formats.timeline import make_json_timeline

        make_json_timeline(
            args.api,
            inp.path,
            output_path,
            obj_sheet,
            chunks,
            args.background,
            log,
        )
        return num_cuts, output_path

    if args.export == "premiere":
        from auto_editor.formats.premiere import premiere_xml

        premiere_xml(inp, temp, output_path, chunks)
        return num_cuts, output_path

    if args.export == "final-cut-pro":
        from auto_editor.formats.final_cut_pro import fcp_xml

        fcp_xml(inp, output_path, chunks)
        return num_cuts, output_path

    if args.export == "shotcut":
        from auto_editor.formats.shotcut import shotcut_xml

        shotcut_xml(inp, output_path, chunks)
        return num_cuts, output_path

    def pad_chunk(
        chunk: Tuple[int, int, float], total_frames: int
    ) -> List[Tuple[int, int, float]]:

        start = []
        end = []
        if chunk[0] != 0:
            start.append((0, chunk[0], 99999.0))

        if chunk[1] != total_frames - 1:
            end.append((chunk[1], total_frames - 1, 99999.0))

        return start + [chunk] + end

    def make_media(
        inp: FileInfo, chunks: List[Tuple[int, int, float]], output_path: str
    ) -> None:
        from auto_editor.utils.video import mux_quality_media
        from auto_editor.render.video import render_av

        if rules.allow_subtitle:
            from auto_editor.render.subtitle import cut_subtitles

            cut_subtitles(ffmpeg, inp, chunks, temp, log)

        if rules.allow_audio:
            from auto_editor.render.audio import make_new_audio

            for t in range(tracks):
                temp_file = os.path.join(temp, f"{t}.wav")
                new_file = os.path.join(temp, f"new{t}.wav")
                make_new_audio(temp_file, new_file, chunks, log, inp.gfps, progress)

                if not os.path.isfile(new_file):
                    log.error("Audio file not created.")

        video_output = []

        if rules.allow_video:
            for v, vid in enumerate(inp.videos):
                if vid.codec not in ("png", "mjpeg", "webp"):
                    out_path, apply_later = render_av(
                        ffmpeg,
                        v,
                        inp,
                        args,
                        chunks,
                        progress,
                        obj_sheet,
                        rules,
                        temp,
                        log,
                    )
                    video_output.append((v, True, out_path, apply_later))
                elif rules.allow_image:
                    out_path = os.path.join(temp, f"{v}.{vid.codec}")
                    # fmt: off
                    ffmpeg.run(["-i", inp.path, "-map", "0:v", "-map", "-0:V",
                        "-c", "copy", out_path])
                    # fmt: on
                    video_output.append((v, False, out_path, False))

        log.conwrite("Writing output file")

        mux_quality_media(
            ffmpeg,
            video_output,
            rules,
            output_path,
            output_container,
            args,
            inp,
            temp,
            log,
        )

    if args.export == "clip-sequence":
        total_frames = chunks[-1][1]
        clip_num = 0
        for chunk in chunks:
            if chunk[2] == 99999:
                continue
            make_media(
                inp,
                pad_chunk(chunk, total_frames),
                append_filename(output_path, f"-{clip_num}"),
            )
            clip_num += 1
    else:
        make_media(inp, chunks, output_path)
    return num_cuts, output_path



import re
from setuptools import setup, find_packages


def pip_version():
    with open("auto_editor/__init__.py") as f:
        version_content = f.read()

    version_match = re.search(
        r"^__version__ = ['\"]([^'\"]*)['\"]", version_content, re.M
    )

    if version_match:
        return version_match.group(1)

    raise ValueError("Unable to find version string.")


with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name="auto-editor",
    version=pip_version(),
    description="Auto-Editor: Effort free video editing!",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="Unlicense",
    url="https://auto-editor.com",
    project_urls={
        "Bug Tracker": "https://github.com/WyattBlue/auto-editor/issues",
        "Source Code": "https://github.com/WyattBlue/auto-editor",
    },
    author="WyattBlue",
    author_email="wyattblue@auto-editor.com",
    keywords="video audio media editor editing processing nonlinear automatic "
    "silence-detect silence-removal silence-speedup motion-detection",
    packages=find_packages(),
    package_data={
        "auto_editor": [
            "help.json",
            "ffmpeg/LICENSE.txt",
            "ffmpeg/Windows/ffmpeg.exe",
            "ffmpeg/Darwin/ffmpeg",
        ],
    },
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        "numpy>=1.21.0",
        "yt-dlp>=2022.1.21",
        "pillow==9.1.0",
        "av==9.2.0",
    ],
    python_requires=">=3.8",
    classifiers=[
        "Topic :: Multimedia :: Sound/Audio",
        "Topic :: Multimedia :: Video",
        "License :: Public Domain",
        "License :: OSI Approved :: The Unlicense (Unlicense)",
        "Environment :: Console",
        "Natural Language :: English",
        "Intended Audience :: End Users/Desktop",
        "Development Status :: 5 - Production/Stable",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
    entry_points={
        "console_scripts": [
            "auto-editor=auto_editor.__main__:main",
            "aedesc=auto_editor.subcommands.desc:main",
            "aeinfo=auto_editor.subcommands.info:main",
            "aesubdump=auto_editor.subcommands.subdump:main",
            "aegrep=auto_editor.subcommands.grep:main",
            "aelevels=auto_editor.subcommands.levels:main",
        ]
    },
)



import io
import numpy
import struct

from typing import Tuple, Optional


PCM = 0x0001
IEEE_FLOAT = 0x0003
EXTENSIBLE = 0xFFFE


def _read_fmt_chunk(
    fid: io.BufferedReader, is_big_endian: bool
) -> Tuple[int, int, int, int, int]:
    if is_big_endian:
        fmt = ">"
    else:
        fmt = "<"

    size: int = struct.unpack(f"{fmt}I", fid.read(4))[0]

    if size < 16:
        raise ValueError("Binary structure of wave file is not compliant")

    res = struct.unpack(f"{fmt}HHIIHH", fid.read(16))
    bytes_read = 16

    format_tag, channels, fs, bytes_per_second, block_align, bit_depth = res

    if format_tag == EXTENSIBLE and size >= 18:
        ext_chunk_size = struct.unpack(f"{fmt}H", fid.read(2))[0]
        bytes_read += 2
        if ext_chunk_size >= 22:
            extensible_chunk_data = fid.read(22)
            bytes_read += 22
            raw_guid = extensible_chunk_data[6:22]

            if is_big_endian:
                tail = b"\x00\x00\x00\x10\x80\x00\x00\xAA\x00\x38\x9B\x71"
            else:
                tail = b"\x00\x00\x10\x00\x80\x00\x00\xAA\x00\x38\x9B\x71"
            if raw_guid.endswith(tail):
                format_tag = struct.unpack(f"{fmt}I", raw_guid[:4])[0]
        else:
            raise ValueError("Binary structure of wave file is not compliant")

    if format_tag not in {PCM, IEEE_FLOAT}:
        raise ValueError(
            f"Encountered unknown format tag: {format_tag:#06x}, while reading fmt chunk."
        )

    # move file pointer to next chunk
    if size > bytes_read:
        fid.read(size - bytes_read)

    # fmt should always be 16, 18 or 40, but handle it just in case
    _handle_pad_byte(fid, size)

    return format_tag, channels, fs, block_align, bit_depth


def _read_data_chunk(
    fid: io.BufferedReader,
    format_tag: int,
    channels: int,
    bit_depth: int,
    is_big_endian: bool,
    block_align: int,
    data_size: Optional[int],
) -> numpy.memmap:

    fmt = ">" if is_big_endian else "<"
    size: int = struct.unpack(f"{fmt}I", fid.read(4))[0]
    if data_size is not None:
        # size is only 32-bits here, so get real size from header.
        size = data_size

    bytes_per_sample = block_align // channels
    # bytes per sample, a.k.a sample container size

    n_samples = size // bytes_per_sample

    if bytes_per_sample in (3, 5, 6, 7):
        raise ValueError(f"Unsupported bytes per sample: {bytes_per_sample}")

    if format_tag == PCM:
        if 1 <= bit_depth <= 8:
            dtype = "u1"  # WAVs of 8-bit integer or less are unsigned
        elif bit_depth <= 64:
            dtype = f"{fmt}i{bytes_per_sample}"
        else:
            raise ValueError(
                f"Unsupported bit depth: the WAV file has {bit_depth}-bit integer data."
            )
    elif format_tag == IEEE_FLOAT:
        if bit_depth in (32, 64):
            dtype = f"{fmt}f{bytes_per_sample}"
        else:
            raise ValueError(
                f"Unsupported bit depth: the WAV file has {bit_depth}-bit floating-point data."
            )
    else:
        raise ValueError(
            f"Unknown wave file format: {format_tag:#06x}. Supported formats: PCM, IEEE_FLOAT"
        )

    start = fid.tell()
    data = numpy.memmap(fid, dtype=dtype, mode="c", offset=start, shape=(n_samples,))
    fid.seek(start + size)

    _handle_pad_byte(fid, size)

    if channels > 1:
        try:
            data = data.reshape(-1, channels)  # type: ignore
        except ValueError:
            data = data[:-1].reshape(-1, channels)  # type: ignore

    return data


def _skip_unknown_chunk(fid: io.BufferedReader, is_big_endian: bool) -> None:
    if is_big_endian:
        fmt = ">I"
    else:
        fmt = "<I"

    data = fid.read(4)
    if data:
        size = struct.unpack(fmt, data)[0]
        fid.seek(size, 1)
        _handle_pad_byte(fid, size)


def _read_rf64_chunk(fid: io.BufferedReader) -> Tuple[int, int, bool]:

    # https://tech.ebu.ch/docs/tech/tech3306v1_0.pdf
    # https://www.itu.int/dms_pubrec/itu-r/rec/bs/R-REC-BS.2088-1-201910-I!!PDF-E.pdf

    heading = fid.read(12)
    if heading != b"\xff\xff\xff\xffWAVEds64":
        raise ValueError(f"Wrong heading: {repr(heading)}")

    chunk_size = fid.read(4)

    bw_size_low = fid.read(4)
    bw_size_high = fid.read(4)

    is_big_endian = bw_size_high > bw_size_low
    fmt = ">I" if is_big_endian else "<I"

    data_size_low = fid.read(4)
    data_size_high = fid.read(4)

    # Combine bw_size and data_size to 64-bit ints

    def combine(a, b) -> int:
        return struct.unpack("<Q", a + b)[0]

    file_size = combine(bw_size_low, bw_size_high)
    data_size = combine(data_size_low, data_size_high)

    chunk_size = struct.unpack(fmt, chunk_size)[0]
    fid.read(40 - chunk_size)  # type: ignore

    return data_size, file_size, is_big_endian


def _read_riff_chunk(sig: bytes, fid: io.BufferedReader) -> Tuple[None, int, bool]:
    if sig == b"RIFF":
        is_big_endian = False
        fmt = "<I"
    else:
        is_big_endian = True
        fmt = ">I"

    file_size: int = struct.unpack(fmt, fid.read(4))[0] + 8

    form = fid.read(4)
    if form != b"WAVE":
        raise ValueError(f"Not a WAV file. RIFF form type is {repr(form)}.")

    return None, file_size, is_big_endian


def _handle_pad_byte(fid: io.BufferedReader, size: int) -> None:
    if size % 2:
        fid.seek(1, 1)


def read(filename: str) -> Tuple[int, numpy.memmap]:
    fid = open(filename, "rb")

    try:
        file_sig = fid.read(4)
        if file_sig in (b"RIFF", b"RIFX"):
            data_size, file_size, is_big_endian = _read_riff_chunk(file_sig, fid)
        elif file_sig == b"RF64":
            data_size, file_size, is_big_endian = _read_rf64_chunk(fid)
        else:
            raise ValueError(f"File format {repr(file_sig)} not supported.")

        fmt_chunk_received = False
        data_chunk_received = False
        while fid.tell() < file_size:

            # read the next chunk
            chunk_id = fid.read(4)

            if not chunk_id:
                if data_chunk_received:
                    # End of file but data successfully read
                    break
                else:
                    raise ValueError("Unexpected end of file.")
            elif len(chunk_id) < 4:
                if fmt_chunk_received and data_chunk_received:
                    pass
                else:
                    raise ValueError(f"Incomplete chunk ID: {repr(chunk_id)}")

            if chunk_id == b"fmt ":
                fmt_chunk_received = True
                format_tag, channels, fs, block_align, bit_depth = _read_fmt_chunk(
                    fid, is_big_endian
                )
            elif chunk_id == b"data":
                data_chunk_received = True
                if not fmt_chunk_received:
                    raise ValueError("No fmt chunk before data")

                data = _read_data_chunk(
                    fid,
                    format_tag,
                    channels,
                    bit_depth,
                    is_big_endian,
                    block_align,
                    data_size,
                )
            else:
                _skip_unknown_chunk(fid, is_big_endian)

    finally:
        fid.seek(0)

    return fs, data
